from numpy import argmin, array, sqrt


def spanning_tree(X, Y):
    """
    Calculate a minimum spanning tree distance for a layout.
    Minimum spanning tree heuristic algorithm from Topfarm0.

    Parameters
    ----------
    x, y: ndarray([n_wt])
            X,Y positions of the wind turbines

    :return: connections: dict of keys (i_wt, j_wt) and value the distance between the two wind turbine index

    the total cable length is simply sum(list(connections.values()))
    """
    n_wt = len(X)
    # X, Y = positions[:,0], positions[:,1]

    def dist(i, j):
        return sqrt((X[i] - X[j])**2.0 + (Y[i] - Y[j])**2.0)

    turblist = range(n_wt)
    connections = {}
    islands = []

    # Look for the smallest connections, and cluster them together.
    for i_wt in turblist:
        not_i_wt = list(filter(lambda x: x != i_wt, turblist))
        distances = sqrt((X[not_i_wt] - X[i_wt])**2.0 + (Y[not_i_wt] - Y[i_wt])**2.0)
        id = argmin(distances)
        closest_wt = not_i_wt[id]

        # Add the connection to the structure
        connections[(i_wt, closest_wt)] = distances[id]
        #   connections{closest_wt}(end+1) = i_wt

        # Add the turbine to an island
        found = False
        for iIS, island in enumerate(islands):
            # Check if it's nearest turbine is already in an island
            if closest_wt in island or i_wt in island:
                # Check if we have already found that the turbine pair is part of an island
                if found:
                    # Those two islands are connected (iIs & id_island)
                    # let's merge them
                    island += filter(lambda x: x != i_wt and x != closest_wt, islands[id_island])
                    del islands[id_island]

                found = True
                id_island = iIS
                # If the current wt is not in the island
                if i_wt not in island:
                    # Add it to the island
                    island.append(i_wt)

                # If the closest_wt is not in the island
                if closest_wt not in island:
                    # Add it to the island
                    island.append(closest_wt)

        # If no island connected to the turbine pair has been found,
        # then we create a new one
        if not found:
            # Create a new island
            islands.append([i_wt, closest_wt])

    # Connect the islands together

    while len(islands) > 1:
        # Look for the closest turbine that is not in the first island
        dist_list = array([[dist(i_wt, j_wt), i_wt, j_wt]
                          for i_wt in islands[0]
                          for j_wt in turblist
                          if j_wt not in islands[0]])
        amin = argmin(dist_list[:, 0])
        i_wt = int(dist_list[amin, 1])
        j_wt = int(dist_list[amin, 2])

        # Add the connection to the structure
        if (i_wt, j_wt) not in connections and (j_wt, i_wt) not in connections:
            connections[(i_wt, j_wt)] = dist_list[amin, 0]
        # if i_wt not in connections[j_wt]:
        #    connections[j_wt].append(i_wt)

        for i, island in enumerate(islands):
            if j_wt in island:
                islands[0] += island
                del islands[i]

    return connections



import os
import threading
import time

import pytest

import numpy as np
from topfarm import TopFarm
from topfarm.cost_models import fuga
from topfarm.tests import uta
from topfarm.plotting import NoPlot
from topfarm.cost_models.cost_model_wrappers import AEPCostModelComponent
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm._topfarm import TopFarmProblem


def check_lib_exists():
    try:
        import py_colonel
    except ModuleNotFoundError:
        pytest.xfail("Colonel submodule not found\n")
    from py_colonel.py_colonel_lib import fugalib_path
    if os.path.isfile(fugalib_path) is False:
        pytest.xfail("Fugalib '%s' not found\n" % fugalib_path)


def _fuga(tb_x=[423974, 424033], tb_y=[6151447, 6150889], wind_atlas='MyFarm/north_pm15_only.lib'):
    check_lib_exists()
    from topfarm.cost_models.fuga.py_fuga import PyFuga, fuga_path
    pyFuga = PyFuga()
    pyFuga.setup(farm_name='Horns Rev 1',
                 turbine_model_path=fuga_path + 'LUTs-T/', turbine_model_name='Vestas_V80_(2_MW_offshore)[h=70.00]',
                 tb_x=tb_x, tb_y=tb_y,
                 mast_position=(0, 0, 70), z0=0.03, zi=400, zeta0=0,
                 farms_dir=fuga_path + 'LUTs-T/Farms/', wind_atlas_path=wind_atlas, climate_interpolation=False)
    return pyFuga


@pytest.fixture
def get_fuga():
    return _fuga


@pytest.fixture
def pyFuga():
    return _fuga()


def _test_parallel(i):
    pyFuga = get_fuga()()
    print(pyFuga.stdout_filename, i)
    for _ in range(1):
        print(threading.current_thread(), i)
        np.testing.assert_array_almost_equal(pyFuga.get_aep(np.array([[0, 0], [0, 200]]).T), [
                                             12.124883, 14.900544, 0.3458, 0.813721])
        time.sleep(1)


def testCheckVersion(get_fuga):
    check_lib_exists()
    from py_colonel.py_colonel_lib import fugalib_path, PascalDLL
    from topfarm.cost_models.fuga.py_fuga import fuga_path
    lib = PascalDLL(fuga_path + "FugaLib/FugaLib.%s" % ('so', 'dll')[os.name == 'nt'])
    with pytest.raises(Exception, match="This version of FugaLib supports interface version "):
        lib.CheckInterfaceVersion(1)
    pyFuga = get_fuga()  # check that current interface version match
    pyFuga.cleanup()


def testSetup(get_fuga):
    pyFuga = get_fuga()
    assert pyFuga.get_no_turbines() == 2
    assert "Loading" in pyFuga.log

    # check that new setup resets number of turbines
    pyFuga = get_fuga()
    assert pyFuga.get_no_turbines() == 2
    pyFuga.cleanup()


def testAEP_one_tb(get_fuga):
    pyFuga = get_fuga([0], [0])
    np.testing.assert_array_almost_equal(pyFuga.get_aep(np.array([[0], [0]]).T), [
                                         8.2896689155874324, 8.2896689155874324, 0.472841, 1.])
    pyFuga.cleanup()


def testAEP(pyFuga):
    np.testing.assert_array_almost_equal(pyFuga.get_aep(np.array([[0, 1000], [0, 0]]).T), [
                                         2 * 8.2896689155874324, 2 * 8.2896689155874324, 0.472841, 1.])
    np.testing.assert_array_almost_equal(pyFuga.get_aep_gradients(np.array([[0, 1000], [0, 0]]).T), [[0, 0],
                                                                                                     [0, 0],
                                                                                                     [1.047718e-002, 9.801237e-003]])
    np.testing.assert_array_almost_equal(pyFuga.get_aep(np.array([[0, 0], [0, 200]]).T), [
                                         14.472994, 16.579338, 0.412768, 0.872954])
    np.testing.assert_array_almost_equal(pyFuga.get_aep_gradients(np.array([[0, 0], [0, 200]]).T), [[0, 0],
                                                                                                    [-7.02042144e-03,
                                                                                                        7.02042144e-03],
                                                                                                    [3.099291e-02, -1.459773e-02]])
    np.testing.assert_array_almost_equal(pyFuga.get_aep(np.array([[0, 200], [0, 200]]).T), [
                                         16.543667, 16.579338, 0.471824, 0.997848])
    np.testing.assert_array_almost_equal(pyFuga.get_aep_gradients(np.array([[0, 200], [0, 200]]).T), [[-1.679974e-05, 1.679974e-05],
                                                                                                      [7.255895e-06, -7.255895e-06],
                                                                                                      [2.002942e-02, 3.759327e-06]])
    pyFuga.cleanup()


def testLargeOffset(pyFuga):
    o = 1.e16
    np.testing.assert_array_almost_equal(pyFuga.get_aep(
        np.array([[0 + o, 0 + o], [0 + o, 200 + o]]).T), [14.472994, 16.579338, 0.412768, 0.872954])
    np.testing.assert_array_almost_equal(pyFuga.get_aep_gradients(), [[0, 0],
                                                                      [-7.02042144e-03, 7.02042144e-03],
                                                                      [3.099291e-02, -1.459773e-02]])


def get_tf(init_pos, pyFuga, boundary, boundary_type='convex_hull'):
    return TopFarmProblem(
        dict(zip('xy', init_pos.T)),
        pyFuga.get_TopFarm_cost_component(),
        constraints=[SpacingConstraint(160),
                     XYBoundaryConstraint(boundary, boundary_type)],
        driver=EasyScipyOptimizeDriver(disp=False))


def testAEP_topfarm(get_fuga):
    pyFuga = get_fuga()
    init_pos = np.array([[0, 0], [1000, 0]])
    tf = get_tf(init_pos, pyFuga, init_pos, 'square')
    cost, _ = tf.evaluate()
    np.testing.assert_array_almost_equal(cost, -16.579337831174865)


def test_pyfuga_cmd():
    check_lib_exists()
    from topfarm.cost_models.fuga.py_fuga import PyFuga
    pyFuga = PyFuga()
    pyFuga.execute(r'echo "ColonelInit"')
    assert pyFuga.log.strip().split("\n")[-1] == 'ColonelInit'


def testAEP_topfarm_optimization_4tb(get_fuga):
    D = 80.0
    B = 3 * D + 10
    init_pos = np.array([(0, 3 * D), (0, D), (0, -D), (0, -3 * D)])
    init_pos[:, 0] += [-80, 0, 0, 80]

    wind_atlas = 'MyFarm/north_pm45_only.lib'
    pyFuga = get_fuga(init_pos[:1, 0], init_pos[:1, 1], wind_atlas=wind_atlas)
    AEP_pr_tb = pyFuga.get_aep()[1]
    pyFuga = get_fuga(init_pos[:, 0], init_pos[:, 1], wind_atlas=wind_atlas)
    boundary = [(-B, B), (B, B), (B, -B), (-B, -B), (-B, B)]

    plot_comp = NoPlot()
    # plot_comp= PlotComp()
    tf = get_tf(init_pos, pyFuga, boundary)
    cost, _, rec = tf.optimize()
    uta.assertAlmostEqual(-cost, AEP_pr_tb * 4, delta=.2)

    tf.plot_comp.show()

#     # Plot wake map
#     f, a, k = lib_reader.read_lib(fuga_path + 'LUTs-T/Farms/' + wind_atlas)
#     wr = WindResource(f, a, k, np.zeros_like(k))
#     pyFuga.plot_wind_field_with_boundary(10, zip(range(360)[::2], wr.f[::2] * 2), 'XY', 70, boundary)


@pytest.mark.parametrize('scale', [(.001),
                                   (1),
                                   (1000)])
def testAEP_topfarm_optimization_2tb_scale(get_fuga, scale):
    D = 80.0
    B = 2 * D + 10
    init_pos = np.array([(-10, 1 * D), (10, - D)])

    wind_atlas = 'MyFarm/north_pm30_only.lib'
    pyFuga = get_fuga(init_pos[:1, 0], init_pos[:1, 1], wind_atlas=wind_atlas)
    AEP_pr_tb = pyFuga.get_aep()[1]
    pyFuga = get_fuga(init_pos[:, 0], init_pos[:, 1], wind_atlas=wind_atlas)
    boundary = [(-B, B), (B, B), (B, -B), (-B, -B), (-B, B)]

    plot_comp = NoPlot()
    # plot_comp = PlotComp()

    cost_comp = AEPCostModelComponent('xy', init_pos.shape[0],
                                      lambda x, y: scale * pyFuga.get_aep(np.array([x, y]).T)[0],  # only aep
                                      lambda x, y: scale * pyFuga.get_aep_gradients(np.array([x, y]).T)[:2])  # only dAEPdx and dAEPdy

    tf = TopFarmProblem(
        dict(zip('xy', init_pos.T)),
        cost_comp,
        constraints=[SpacingConstraint(2 * D),
                     XYBoundaryConstraint(boundary)],
        plot_comp=plot_comp,
        driver=EasyScipyOptimizeDriver(tol=1e-8, disp=False),
        expected_cost=AEP_pr_tb * 2 * scale)
    cost, _, rec = tf.optimize()
    tf.plot_comp.show()
    uta.assertAlmostEqual(-cost / scale, AEP_pr_tb * 2, delta=.02)


#     # Plot wake map
#     f, a, k = lib_reader.read_lib(fuga_path + 'LUTs-T/Farms/' + wind_atlas)
#     wr = WindResource(f, a, k, np.zeros_like(k))
#     #pyFuga.plot_wind_field_with_boundary(10, zip(range(360)[::2], wr.f[::2] * 2), 'XY', 70, boundary)
#     pyFuga.plot_wind_field_with_boundary(10, 0, 'XY', 70, boundary)

# @pytest.mark.xfail
# def test_parallel():
#     from multiprocessing import Pool
#     with Pool(5) as p:
#         print(p.map(_test_parallel, [1, 2]))



from openmdao.api import SqliteRecorder
import numpy as np
import os
import pickle
import topfarm
from copy import deepcopy
from openmdao.core.driver import Driver
from openmdao.core.system import System


def split_record_id(record_id):
    if record_id is None:
        return "", ""
    folder, record_id = os.path.split(record_id)
    filename, load_case, *_ = (record_id + ":latest").split(":")
    if load_case == "":
        load_case = 'latest'
    return os.path.join(folder, filename).replace("\\", "/"), load_case


def recordid2filename(record_id):
    if record_id is None:
        return "", ""
    filename, load_case = split_record_id(record_id)

    if not filename.lower().endswith(".pkl"):
        filename += ".pkl"
    folder, filename = os.path.split(filename)
    if folder == "":
        folder = "recordings"
    return os.path.join(folder, filename).replace("\\", "/"), load_case.lower()


class TopFarmListRecorder(SqliteRecorder):
    def __init__(self, record_id=None, filepath='cases.sql', append=False, pickle_version=2, record_viewer_data=False):
        super().__init__(filepath, append, pickle_version, record_viewer_data)
        self.iteration_coordinate_lst = []
        self.filepath = filepath
        self.driver_iteration_dict = {}
        self.meta_field_names = ['counter', 'iteration_coordinate', 'timestamp', 'success', 'msg']
        self._abs2prom = {'input': {}, 'output': {}}
        self._prom2abs = {'input': {}, 'output': {}}
        self._abs2meta = {}
        filepath, _ = recordid2filename(record_id)
        self.load_if_exists(record_id)

    def get(self, key):
        if isinstance(key, (tuple, list)):
            return np.array([self.get(k) for k in key]).T
        res = np.array(self.driver_iteration_dict[key])
        if len(res.shape) > 1 and res.shape[-1] == 1:
            res = np.squeeze(res, -1)
        return res

    def __getitem__(self, key):
        return self.get(key)

    @property
    def num_cases(self):
        return len(self.iteration_coordinate_lst)

    def record_iteration_driver(self, recording_requester, data, metadata):
        """
        Record data and metadata from a Driver.

        Parameters
        ----------
        recording_requester : object
            Driver in need of recording.
        data : dict
            Dictionary containing desvars, objectives, constraints, responses, and System vars.
        metadata : dict
            Dictionary containing execution metadata.
        """
        self.iteration_coordinate_lst.append(self._iteration_coordinate)
        meta_fields = [('counter', self._counter),
                       ('iteration_coordinate', self._iteration_coordinate),
                       ('timestamp', metadata['timestamp']),
                       ('success', metadata['success']),
                       ('msg', metadata['msg'])]

        out_keys = []
        in_keys = []
        if self.num_cases == 1:
            for key in data['out']:
                rec_key = key.split('.')[-1]
                out_keys.append(rec_key)
                self.driver_iteration_dict[rec_key] = [data['out'][key].copy()]
            for key in data['in']:
                rec_key = key.split('.')[-1]
                if rec_key not in out_keys:
                    in_keys.append(rec_key)
                    self.driver_iteration_dict[rec_key] = [data['in'][key].copy()]
            for k, v in meta_fields:
                self.driver_iteration_dict[k] = [v]

        else:
            for key in data['out']:
                rec_key = key.split('.')[-1]
                self.driver_iteration_dict[rec_key].append(data['out'][key].copy())
            for key in data['in']:
                rec_key = key.split('.')[-1]
                if rec_key in in_keys:
                    self.driver_iteration_dict[rec_key].append(data['in'][key].copy())
            for k, v in meta_fields:
                self.driver_iteration_dict[k].append(v)

    def _initialize_database(self):
        pass

    def animate_turbineXY(self, duration=10, tail=5, plot_initial=True, filename=None):
        import matplotlib.pyplot as plt
        x, y = self[topfarm.x_key], self[topfarm.y_key]
        cost = self.get('cost')
        boundary = self['xy_boundary'][-1]
        boundary = np.r_[boundary, boundary[:1]]
        n_wt = x.shape[1]
        from matplotlib import animation

        # color_cycle = iter(matplotlib.rcParams['axes.prop_cycle'])
        colors = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
                  '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']
        colors = colors * (n_wt // 10 + 1)

        fig, ax = plt.subplots()
        ln = [plt.plot([], [], '.-', color=c, animated=True)[0] for c in colors]
        ln += [plt.plot([], [], '--', color=c, animated=True)[0] for c in colors]
        ln += [plt.plot([], [], 'xk', animated=True)[0] for _ in colors]
        title = ax.text(0.5, .95, "", bbox={'facecolor': 'w', 'alpha': .8, 'pad': 5},
                        transform=ax.transAxes, ha="center")
        ln += title,

        def init():
            if len(boundary) > 0:
                plt.plot(boundary[:, 0], boundary[:, 1], 'k')
            else:
                ax.set_xlim([np.min(x), np.max(x)])
                ax.set_ylim([np.min(y), np.max(y)])
            plt.axis('equal')
            return ln

        init()

        def update(frame):
            title.set_text("%f (%.2f%%)" % (cost[frame],
                                            (cost[frame] - cost[0]) / cost[0] * 100))
            for i in range(n_wt):
                ln[i].set_data(x[max(0, frame - tail):frame + 1, i], y[max(0, frame - tail):frame + 1, i])
                if plot_initial:
                    ln[i + n_wt].set_data(np.r_[x[0, i], x[frame, i]], np.r_[y[0, i], y[frame, i]])
                ln[i + 2 * n_wt].set_data(x[frame, i], y[frame, i])

            return ln

        ani = animation.FuncAnimation(fig, update, frames=len(x),
                                      init_func=init, blit=True, interval=duration / len(x) * 1000, repeat=False)
        if filename:
            Writer = animation.writers['ffmpeg']
            writer = Writer(fps=len(x) / duration, metadata=dict(artist='Me'), bitrate=1800)
            os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
            ani.save(filename, writer=writer)
        else:
            plt.show()

    def split_record_id(self, record_id):
        return split_record_id(record_id)

    def recordid2filename(self, record_id):
        return recordid2filename(record_id)

    def save(self, record_id=None):
        record_id = record_id or self.filename
        filename, _ = self.recordid2filename(record_id)
        self.list_save(filename)

    def list_save(self, filename):
        d = {'driver_iteration_dict': self.driver_iteration_dict,
             'iteration_coordinate_lst': self.iteration_coordinate_lst,
             '_abs2prom': self._abs2prom,
             '_prom2abs': self._prom2abs,
             '_abs2meta': self._abs2meta,
             }

        if os.path.dirname(filename) != "":
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'wb') as fid:
            pickle.dump([self.__class__, d], fid)

    def load_if_exists(self, record_id):
        self.filename = self.recordid2filename(record_id)[0]
        if self.filename and os.path.isfile(self.filename):
            self.load(record_id)
        return self

    def load(self, record_id):
        filename, load_case = self.recordid2filename(record_id)
        if not filename or os.path.isfile(filename) is False:
            raise FileNotFoundError("No such file '%s'" % filename)

        if load_case == 'none' or load_case == '0':
            return self

        self.list_load(filename)
        if load_case == 'latest':
            load_case = None
        elif load_case == 'best':
            load_case = np.argmin(self.get('aggr_cost')) + 1
        else:
            load_case = int(load_case)
        self.driver_iteration_dict = {k: v[:load_case] for k, v in self.driver_iteration_dict.items()}
        self.iteration_coordinate_lst = self.iteration_coordinate_lst[:load_case]

        self.filename, self.load_case = filename, load_case
        return self

    def list_load(self, filename):
        if not filename or os.path.isfile(filename) is False:
            raise FileNotFoundError("No such file '%s'" % filename)
        with open(filename, 'rb') as fid:
            cls, attributes = pickle.load(fid)
        assert self.__class__ == cls
        self.__dict__.update(attributes)
        return self

    def keys(self):
        return list(self.driver_iteration_dict)


class NestedTopFarmListRecorder(TopFarmListRecorder):
    def __init__(self, nested_comp, record_id=None):
        TopFarmListRecorder.__init__(self, record_id)
        self.nested_comp = nested_comp

    def record_iteration_driver(self, recording_requester, data, metadata):
        """
        Record data and metadata from a Driver.

        Parameters
        ----------
        recording_requester : object
            Driver in need of recording.
        data : dict
            Dictionary containing desvars, objectives, constraints, responses, and System vars.
        metadata : dict
            Dictionary containing execution metadata.
        """
        recorder = getattr(self.nested_comp.problem, 'recorder', None)
        if self.num_cases == 0:
            self.driver_iteration_dict['recorder'] = []
        TopFarmListRecorder.record_iteration_driver(self, recording_requester, data, metadata)

        self.driver_iteration_dict['recorder'].append(recorder)



from py_wake.aep_calculator import AEPCalculator
from py_wake.examples.data.iea37._iea37 import IEA37_WindTurbines, IEA37Site
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.plotting import XYPlotComp
from topfarm.cost_models.cost_model_wrappers import AEPCostModelComponent
from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle
from py_wake.wake_models.gaussian import IEA37SimpleBastankhahGaussian
import topfarm
import numpy as np
from scipy.interpolate.interpolate import RegularGridInterpolator


class PyWakeAEP(AEPCalculator):
    """TOPFARM wrapper for PyWake AEP calculator"""

    def get_TopFarm_cost_component(self, n_wt, wd=None, ws=None):
        """Create topfarm-style cost component

        Parameters
        ----------
        n_wt : int
            Number of wind turbines
        """
        return AEPCostModelComponent(
            input_keys=['x', 'y'],
            n_wt=n_wt,
            cost_function=lambda **kwargs:
                self.calculate_AEP(x_i=kwargs[topfarm.x_key],
                                   y_i=kwargs[topfarm.y_key],
                                   h_i=kwargs.get(topfarm.z_key, None),
                                   type_i=kwargs.get(topfarm.type_key, None),
                                   wd=wd, ws=ws).sum(),
            output_unit='GWh')

    def get_aep4smart_start(self, ws=[6, 8, 10, 12, 14], wd=np.arange(360)):
        def aep4smart_start(X, Y, wt_x, wt_y):
            x = np.sort(np.unique(X))
            y = np.sort(np.unique(Y))
            X_j, Y_j, aep_map = self.aep_map(x, y, 0, wt_x, wt_y, wd=wd, ws=ws)
            return RegularGridInterpolator((y, x), aep_map)(np.array([Y, X]).T)
        return aep4smart_start


def main():
    if __name__ == '__main__':
        site = IEA37Site(16)
        windTurbines = IEA37_WindTurbines()
        wake_model = IEA37SimpleBastankhahGaussian(site, windTurbines)
        aep_calc = PyWakeAEP(wake_model)
        tf = TopFarmProblem(
            design_vars=dict(zip('xy', site.initial_position.T)),
            cost_comp=aep_calc.get_TopFarm_cost_component(16),
            driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbinePosition_Circle(), max_iter=5),
            constraints=[CircleBoundaryConstraint([0, 0], 1300.1)],
            plot_comp=XYPlotComp())
        tf.optimize()
        tf.plot_comp.show()


main()



from openmdao.drivers.doe_driver import DOEDriver
from openmdao.drivers.doe_generators import ListGenerator, UniformGenerator
import pytest
import numpy as np
from topfarm.cost_models.dummy import DummyCost
from topfarm.tests import npt, uta
from topfarm.constraint_components.constrained_generator import ConstrainedXYZGenerator,\
    ConstrainedGenerator, ConstrainedDiscardGenerator
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm._topfarm import TopFarmProblem


def get_InitialXYZOptimizationProblem(driver,
                                      min_spacing=None,
                                      turbineXYZ=[[0, 0, 0],
                                                  [2, 2, 2]],
                                      xy_boundary=[(10, 6), (11, 8)],
                                      xy_boundary_type='rectangle',
                                      z_boundary=[3, 4]):
    cost_comp = DummyCost([(1, 0, 4),
                           (0, 1, 3)], ['x', 'y', 'z'])
    turbineXYZ = np.array(turbineXYZ)
    desvar = dict(zip('xy', turbineXYZ.T))
    desvar['z'] = (turbineXYZ[:, 2], z_boundary[0], z_boundary[1])
    constraints = [XYBoundaryConstraint(xy_boundary, xy_boundary_type)]
    if min_spacing:
        constraints.append(SpacingConstraint(min_spacing))

    return TopFarmProblem(
        desvar,
        cost_comp,
        constraints=constraints,
        driver=driver)


@pytest.fixture
def get_tf():
    return get_InitialXYZOptimizationProblem


def DOE_lst():
    return get_InitialXYZOptimizationProblem(xy_boundary=[(0, 0), (10, 0), (10, 10)],
                                             xy_boundary_type='convex_hull',
                                             driver=UniformGenerator(10, seed=0)).get_DOE_list()


@pytest.mark.parametrize('driver', [ConstrainedGenerator(UniformGenerator(10, seed=0)),
                                    ConstrainedDiscardGenerator(UniformGenerator(10, seed=0)),
                                    ConstrainedGenerator(DOE_lst()),
                                    ConstrainedDiscardGenerator(DOE_lst())])
def test_with_convex_boundary(get_tf, driver):
    tf = get_tf(xy_boundary=[(0, 0), (10, 0), (10, 10)], xy_boundary_type='convex_hull',
                driver=driver)
    arr = tf.get_DOE_array()
    x, y = [arr[:, i] for i in range(2)]
    uta.assertGreaterEqual(x.min(), 0)  # x
    uta.assertGreaterEqual(y.min(), 0)  # y
    npt.assert_array_less(y, x)


@pytest.mark.parametrize('driver', [ConstrainedGenerator(UniformGenerator(10, 0)),
                                    ConstrainedDiscardGenerator(UniformGenerator(10, 0))])
def test_with_constrained_generator_polygon(get_tf, driver):
    tf = get_tf(xy_boundary=[(0, 0), (10, 0), (10, 10)],
                xy_boundary_type='polygon',
                driver=driver)
    arr = tf.get_DOE_array()
    x, y = [arr[:, i] for i in range(2)]
    uta.assertGreaterEqual(x.min(), 0)  # x
    uta.assertGreaterEqual(y.min(), 0)  # y
    npt.assert_array_less(y, x)


def test_with_constrained_generator_spacing(get_tf):
    lst = [[('x', [1, 1]), ('y', [3, 4]), ('z', [5, 6])],
           [('x', [2, 2]), ('y', [6, 5]), ('z', [2, 1])]]

    tf = get_tf(xy_boundary=[(0, 0), (10, 10)],
                xy_boundary_type='rectangle',
                min_spacing=2,
                driver=ConstrainedXYZGenerator(lst))
    arr = tf.get_DOE_array()

    x, y = [arr[:, i] for i in range(2)]
    assert all(np.sqrt(np.diff(x)**2 + np.diff(y)**2) >= 2)


@pytest.mark.parametrize('driver', [ConstrainedGenerator(UniformGenerator(10, 0)),
                                    ConstrainedDiscardGenerator(UniformGenerator(10, 0))])
def test_with_convex_boundary_and_constrain(get_tf, driver):
    tf = get_tf(xy_boundary=[(0, 0), (10, 0), (10, 10)], xy_boundary_type='convex_hull',
                min_spacing=2, driver=driver)
    arr = tf.get_DOE_array()
    x, y = [arr[:, i] for i in range(2)]
    uta.assertGreaterEqual(x.min(), 0)  # x
    uta.assertGreaterEqual(y.min(), 0)  # y
    npt.assert_array_less(y, x)
    assert all(np.sqrt(np.diff(x)**2 + np.diff(y)**2) >= 2)


if __name__ == '__main__':
    tf = get_InitialXYZOptimizationProblem(xy_boundary=[(0, 0), (10, 0), (10, 10)], xy_boundary_type='convex_hull',
                                           min_spacing=2, driver=ConstrainedDiscardGenerator(UniformGenerator(10, 0)))
    arr = tf.get_DOE_array()
    x, y = [arr[:, i] for i in range(2)]
    uta.assertGreaterEqual(x.min(), 0)  # x
    uta.assertGreaterEqual(y.min(), 0)  # y
    npt.assert_array_less(y, x)



import numpy as np
from topfarm.constraint_components import Constraint, ConstraintComponent
import topfarm


class CapacityConstraint(Constraint):
    def __init__(self, max_capacity=500, rated_power_array=[3000]):
        """Initialize CapacityConstraint

        Parameters
        ----------
        max_capacity : int or float
            Maximum wind farm installed capacity [MW]
        rated_power_array : list
            Rated power list corresponding to turbine type order [kW]
        """
        self.max_capacity = max_capacity
        self.rated_power_array = np.array(rated_power_array) / 1e3
        self.const_id = 'capacity_comp_{}'.format(int(max_capacity))

    @property
    def constraintComponent(self):
        return self.capacity_comp

    def _setup(self, problem):
        self.n_wt = problem.n_wt
        self.capacity_comp = CapacityComp(self.n_wt, self.max_capacity, self.rated_power_array, self.const_id)
        problem.model.pre_constraints.add_subsystem(self.const_id, self.capacity_comp, promotes=[
            topfarm.type_key, 'penalty_' + self.const_id, 'totalcapacity'])

    def setup_as_constraint(self, problem):
        self._setup(problem)
        problem.model.add_constraint('totalcapacity', upper=self.max_capacity)

    def setup_as_penalty(self, problem):
        self._setup(problem)


class CapacityComp(ConstraintComponent):
    """
    Calculates total installed capacity of the wind farm.
    """

    def __init__(self, n_wt, max_capacity, rated_power_array, const_id=None):
        super().__init__()
        self.n_wt = n_wt
        self.max_capacity = max_capacity
        self.rated_power_array = np.array(rated_power_array)
        self.const_id = const_id

    def setup(self):
        self.add_input(topfarm.type_key, np.zeros(self.n_wt, dtype=np.int))
        self.add_output('penalty_' + self.const_id, val=0.0)
        self.add_output('totalcapacity', val=0.0,
                        desc='wind farm installed capacity')
        # Partial declaration is not needed for type or penalty in this case.
        # Because it is only used for integer optimization.

    def compute(self, inputs, outputs):
        outputs['totalcapacity'] = np.sum(self.rated_power_array[inputs[topfarm.type_key].astype(int)])
        outputs['penalty_' + self.const_id] = np.maximum(0, outputs['totalcapacity'][0] - self.max_capacity)

    def satisfy(self):
        pass



from ._topfarm import *
from .deprectated_topfarm_problems import *
import pkg_resources

plugins = {
    entry_point.name: entry_point.load()
    for entry_point
    in pkg_resources.iter_entry_points('topfarm.plugins')
}

__version__ = '2.2.1'
__release__ = '2.2.1'


x_key = 'x'
y_key = 'y'
z_key = 'z'
type_key = 'type'
cost_key = 'cost'



import numpy as np
from topfarm import TopFarm
import pytest
from topfarm.tests import npt
from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
from topfarm.constraint_components.boundary import XYBoundaryConstraint,\
    ConvexBoundaryComp
from topfarm._topfarm import TopFarmProblem

from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
from topfarm.easy_drivers import EasyScipyOptimizeDriver


def testSquare():
    boundary = [(0, 0), (1, 3)]
    b = ConvexBoundaryComp(0, boundary, 'square')
    np.testing.assert_array_equal(b.xy_boundary, [[-1, 0],
                                                  [2, 0],
                                                  [2, 3],
                                                  [-1, 3],
                                                  [-1, 0]])


def testRectangle():
    boundary = [(0, 0), (1, 3)]
    b = ConvexBoundaryComp(0, boundary, 'rectangle')
    np.testing.assert_array_equal(b.xy_boundary, [[0, 0],
                                                  [1, 0],
                                                  [1, 3],
                                                  [0, 3],
                                                  [0, 0]])


def testConvexHull():
    boundary = [(0, 0), (1, 1), (2, 0), (2, 2), (0, 2)]
    b = ConvexBoundaryComp(0, boundary, 'convex_hull')
    np.testing.assert_array_equal(b.xy_boundary, [[0, 0],
                                                  [2, 0],
                                                  [2, 2],
                                                  [0, 2],
                                                  [0, 0]])


def testNotImplemented():
    boundary = [(0, 0), (1, 1), (2, 0), (2, 2), (0, 2)]
    with pytest.raises(NotImplementedError, match="Boundary type 'Something' is not implemented"):
        ConvexBoundaryComp(0, boundary, boundary_type='Something')


def test_z_boundary():
    optimal = np.array([(0, 0, 0)]).T
    tf = TopFarmProblem(
        {'z': (optimal, 70, 90)},
        DummyCost(optimal, 'z'),
        driver=SimpleGADriver())

    desvars = tf.driver._designvars
    np.testing.assert_array_equal(desvars['indeps.z']['lower'], [70, 70])
    np.testing.assert_array_equal(desvars['indeps.z']['upper'], [90, 90])


def test_xyz_boundary():
    optimal = np.array([(0, 0, 0)])
    boundary = [(0, 0), (1, 3)]
    desvar = dict(zip('xy', optimal.T))
    desvar['z'] = (optimal[:, 2], 70, 90)
    b = XYBoundaryConstraint(boundary, boundary_type='rectangle')
    tf = TopFarmProblem(
        desvar,
        DummyCost(optimal, 'xyz'),
        constraints=[b],
        driver=SimpleGADriver())

    np.testing.assert_array_equal(b.constraintComponent.xy_boundary, [[0, 0],
                                                                      [1, 0],
                                                                      [1, 3],
                                                                      [0, 3],
                                                                      [0, 0]])
    desvars = tf.driver._designvars
    np.testing.assert_array_equal(desvars['indeps.z']['lower'], [70])
    np.testing.assert_array_equal(desvars['indeps.z']['upper'], [90])


def test_move_inside():
    pbc = ConvexBoundaryComp(1, [(0, 0), (9, 0), (10, 1), (10, 10)])
    x0, y0 = [3, 3, 3, 12, 12, 12], [3, 5, 10, 8, 10, 12]
    state = pbc.satisfy({'x': x0, 'y': y0})
    x, y = state['x'], state['y']
    if 0:
        import matplotlib.pyplot as plt
        b = np.r_[pbc.xy_boundary, pbc.xy_boundary[:1]]
        plt.plot(b[:, 0], b[:, 1], 'k')
        for x0_, x_, y0_, y_ in zip(x0, x, y0, y):
            plt.plot([x0_, x_], [y0_, y_], '.-')
        plt.show()
    eps = 1e-10
    npt.assert_array_less(y, x + eps)
    npt.assert_array_less(x, 10 + eps)
    npt.assert_array_less(y, 10 + eps)
    npt.assert_array_less(-x, 0 + eps)
    npt.assert_array_less(-y, 0 + eps)



import importlib
import os
import pkgutil
import warnings
import mock
import pytest
import matplotlib.pyplot as plt
import sys
from examples import docs
import subprocess


def get_main_modules():
    package = docs
    modules = []
    for _, modname, _ in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if 'Colonel' in modname:
                continue
            m = importlib.import_module(modname)

        if 'main' in dir(m):
            modules.append(m)
    return modules


def test_print_main_modules():
    print("\n".join([m.__name__ for m in get_main_modules()]))


@pytest.mark.parametrize("module", get_main_modules())
def test_main(module):
    # check that all main module examples run without errors
    #    if os.name == 'posix' and "DISPLAY" not in os.environ:
    #        pytest.xfail("No display")

    def no_show(*args, **kwargs):
        pass
    plt.show = no_show  # disable plt show that requires the user to close the plot

    def no_print(*args, **kwargs):
        pass
    try:
        with mock.patch.object(module, "__name__", "__main__"):
            with mock.patch.object(module, "print", no_print):
                try:
                    from mpi4py import MPI
                    if hasattr(module, 'N_PROCS'):
                        N_PROCS = getattr(module, 'N_PROCS')
                        use_mpi = N_PROCS > 1
                    else:
                        use_mpi = False
                except ImportError:
                    use_mpi = False
                if use_mpi:
                    path = str(module.__spec__.origin)
                    args = ['mpirun', '--allow-run-as-root', '-n', str(N_PROCS), 'python', path]
                    env = os.environ.copy()
                    process = subprocess.Popen(args,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               universal_newlines=True,
                                               env=env)
                    stdout, stderr = process.communicate()
                    if process.returncode != 0:
                        raise EnvironmentError("%s\n%s" % (stdout, stderr))
                else:
                    getattr(module, 'main')()
    except Exception as e:
        raise type(e)(str(e) + ' in %s.main' % module.__name__).with_traceback(sys.exc_info()[2])


if __name__ == '__main__':
    test_main(get_main_modules()[-1])



# from topfarm.constraint_components.boundary_component import BoundaryComp
# from topfarm.constraint_components.spacing_component import SpacingComp
# from topfarm.plotting import PlotComp
import sys
import time
import warnings

from openmdao.api import ScipyOptimizeDriver
import numpy as np
import topfarm
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.boundary_component import PolygonBoundaryComp
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.drivers import random_search_driver
from topfarm.easy_drivers import EasyRandomSearchDriver


def set_deprecated_keys():
    topfarm.x_key = 'turbineX'
    topfarm.y_key = 'turbineY'
    topfarm.z_key = 'turbineZ'
    topfarm.type_key = 'turbineType'


class TurbineTypeOptimizationProblem(TopFarmProblem):
    def __init__(self, cost_comp, turbineTypes, lower, upper, **kwargs):
        sys.stderr.write("%s is deprecated. Use TopFarmProblem instead\n" % self.__class__.__name__)

        TopFarmProblem.__init__(self,
                                design_vars={topfarm.type_key: (turbineTypes, lower, upper)},
                                cost_comp=cost_comp,
                                **kwargs)
        self.setup()


class TurbineXYZOptimizationProblem(TopFarmProblem):
    def __init__(self, cost_comp, turbineXYZ, boundary_comp, min_spacing=None,
                 driver=ScipyOptimizeDriver(), plot_comp=None, record_id=None, expected_cost=1):
        sys.stderr.write("%s is deprecated. Use TopFarmProblem instead\n" % self.__class__.__name__)
        if plot_comp:
            if plot_comp == "default":
                plot_comp = PlotComp()
        turbineXYZ = np.asarray(turbineXYZ)
        design_vars = {xy: v for xy, v in zip([topfarm.x_key, topfarm.y_key], turbineXYZ.T)}
        constraints = []
        if min_spacing:
            constraints.append(SpacingConstraint(min_spacing))

        if isinstance(boundary_comp, PolygonBoundaryComp):
            constraints.append(XYBoundaryConstraint(boundary_comp.xy_boundary, 'polygon'))
        elif len(boundary_comp.xy_boundary):
            constraints.append(XYBoundaryConstraint(boundary_comp.xy_boundary, boundary_comp.boundary_type))

        if turbineXYZ.shape[1] == 3:
            if len(boundary_comp.z_boundary):
                design_vars[topfarm.z_key] = (turbineXYZ[:, 2], boundary_comp.z_boundary[:, 0], boundary_comp.z_boundary[:, 1])
            else:
                design_vars[topfarm.z_key] = turbineXYZ[:, 2]

        TopFarmProblem.__init__(
            self,
            design_vars=design_vars,
            cost_comp=cost_comp,
            driver=driver,
            constraints=constraints,
            plot_comp=plot_comp,
            record_id=record_id,
            expected_cost=expected_cost)
        self.setup()


class TurbineTypeXYZOptimizationProblem(TurbineTypeOptimizationProblem, TurbineXYZOptimizationProblem):
    def __init__(self, cost_comp, turbineTypes, lower, upper, turbineXYZ, boundary_comp, min_spacing=None,
                 driver=EasyRandomSearchDriver(random_search_driver.RandomizeTurbineTypeAndPosition()), plot_comp=None, record_id=None, expected_cost=1):
        sys.stderr.write("%s is deprecated. Use TopFarmProblem instead\n" % self.__class__.__name__)
        TopFarmProblem.__init__(self, cost_comp, driver, plot_comp, record_id, expected_cost)
        TurbineTypeOptimizationProblem.initialize(self, turbineTypes, lower, upper)
        TurbineXYZOptimizationProblem.initialize(self, turbineXYZ, boundary_comp, min_spacing)
        self.setup(check=True, mode=self.mode)

#     @property
#     def state(self):
#         state = {k: self[k] for k in ['turbineX', 'turbineY', 'turbineZ']}
#         state['turbineType'] = self['turbineType'].astype(np.int)
#         state.update(TopFarmProblem.state.fget(self))
#         return state


class InitialXYZOptimizationProblem(TurbineXYZOptimizationProblem):
    def __init__(self, cost_comp, turbineXYZ, boundary_comp=None, min_spacing=None,
                 driver=None, plot_comp=None):
        #          if driver is None:
        #              driver = DOEDriver(shuffle_generator(self, 10))
        sys.stderr.write("%s is deprecated. Use TopFarmProblem instead\n" % self.__class__.__name__)
        TurbineXYZOptimizationProblem.__init__(self, cost_comp, turbineXYZ,
                                               boundary_comp, min_spacing,
                                               driver=driver, plot_comp=plot_comp)
        self.setup()

#     def shuffle_positions(self, shuffle_type='rel', n_iter=1000,
#                           step_size=0.1, pad=1.1, offset=5, plot=False,
#                           verbose=False):
#         if shuffle_type is not None:
#             turbines = spos(self.boundary, self.n_wt, self.min_spacing,
#                             self.turbine_positions, shuffle_type, n_iter,
#                             step_size, pad, offset, plot, verbose)
#             self.problem['turbineX'] = turbines.T[0]
#             self.problem['turbineY'] = turbines.T[1]


class TopFarm(TopFarmProblem):
    """Wrapper of TurbineXYZOPtimizationProblem for backward compatibility
    """

    def __init__(self, turbines, cost_comp, min_spacing, boundary,
                 boundary_type='convex_hull', plot_comp=None,
                 driver=ScipyOptimizeDriver(),
                 record_id="Opt_%s" % time.strftime("%Y%m%d_%H%M%S"),
                 expected_cost=1):
        sys.stderr.write("%s is deprecated. Use TopFarmProblem instead\n" % self.__class__.__name__)
        constraints = []
        if min_spacing and len(turbines) > 1:
            constraints.append(SpacingConstraint(min_spacing))
        if boundary is not None:
            constraints.append(XYBoundaryConstraint(boundary, boundary_type))
        TopFarmProblem.__init__(
            self,
            design_vars={k: v for k, v in zip([topfarm.x_key, topfarm.y_key, topfarm.z_key], np.asarray(turbines).T)},
            cost_comp=cost_comp,
            driver=driver,
            constraints=constraints,
            plot_comp=plot_comp,
            record_id=record_id,
            expected_cost=expected_cost)
        self.setup()



from openmdao.drivers.doe_generators import UniformGenerator
import pytest
from topfarm.cost_models.dummy import DummyCost
from topfarm.tests import npt
from topfarm.parallel_runner import ParallelRunner
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm import TopFarmProblem


@pytest.fixture("module")
def parallelRunner():
    return ParallelRunner()


def get_InitialXYZOptimizationProblem(driver):
    return TopFarmProblem(
        {'x': [0, 2], 'y': [0, 2], 'z': ([0, 2], 3, 4)},
        cost_comp=DummyCost([(1, 0, 4),
                             (0, 1, 3)], 'xyz'),
        constraints=[XYBoundaryConstraint([(10, 6), (11, 8)], 'rectangle')],
        driver=driver)


@pytest.fixture
def lst():
    return get_InitialXYZOptimizationProblem(driver=UniformGenerator(200)).get_DOE_list()


def seq_runner_example(lst):
    return get_InitialXYZOptimizationProblem(lst).optimize()


def test_parallel_run(lst, parallelRunner):
    # run sequential
    s_cost, s_state, s_recorder = seq_runner_example(lst)

    # run in parallel
    (p_cost, p_state, p_recorder), results = parallelRunner(lst, seq_runner_example)
    npt.assert_equal(s_cost, p_cost)



import pytest

import numpy as np

from topfarm.cost_models.dummy import DummyCost
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle, RandomizeTurbinePosition_Square,\
    RandomizeTurbineTypeAndPosition, RandomizeTurbinePosition_Normal,\
    RandomizeAllUniform, RandomizeAllRelativeMaxStep
from topfarm.easy_drivers import EasyScipyOptimizeDriver, EasySimpleGADriver, EasyRandomSearchDriver, EasyPyOptSparseSNOPT, EasyPyOptSparseIPOPT
from topfarm.plotting import NoPlot
from topfarm.tests import uta, npt
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm._topfarm import TopFarmProblem
from topfarm.cost_models.cost_model_wrappers import CostModelComponent


initial = np.array([[6, 0, 0], [6, -8, 0], [1, 1, 0]])  # initial turbine layouts
optimal = np.array([[2.5, -3, 1], [6, -7, 2], [4.5, -3, 3]])  # optimal turbine layouts
boundary = np.array([(0, 0), (6, 0), (6, -10), (0, -10)])  # turbine boundaries
desired = np.array([[3, -3, 1], [7, -7, 2], [4, -3, 4]])  # desired turbine layouts


@pytest.fixture
def topfarm_generator_scalable():
    def _topfarm_obj(driver, xy_scale=[1, 1], cost_scale=1, cost_offset=0, spacing=2):
        # from topfarm.cost_models.dummy import DummyCostPlotComp
        # plot_comp = DummyCostPlotComp(desired[:, :2] * xy_scale, plot_improvements_only=False)
        plot_comp = NoPlot()

        class DummyCostScaled(DummyCost):
            def cost(self, **kwargs):
                opt = self.optimal_state
                return np.sum([(kwargs[n] - opt[:, i])**2 for i, n in enumerate(self.input_keys)]) * cost_scale + cost_offset

            def grad(self, **kwargs):
                opt = self.optimal_state
                return [(2 * cost_scale * (kwargs[n] - opt[:, i])) for i, n in enumerate(self.input_keys)]

        return TopFarmProblem(
            dict(zip('xy', (initial[:, :2] * xy_scale).T)),
            DummyCostScaled(desired[:, :2] * xy_scale),
            constraints=[SpacingConstraint(spacing * xy_scale[0]),
                         XYBoundaryConstraint(boundary * xy_scale)],
            driver=driver,
            plot_comp=plot_comp,
            expected_cost=1.5 * cost_scale)
    return _topfarm_obj


@pytest.fixture
def topfarm_generator():
    def _topfarm_obj(driver, spacing=2, keys='xy'):
        # from topfarm.cost_models.dummy import DummyCostPlotComp
        # plot_comp = DummyCostPlotComp(desired[:,:len(keys)], plot_improvements_only=True)
        plot_comp = NoPlot()

        return TopFarmProblem(
            dict(zip(keys, initial.T[:len(keys)])),
            DummyCost(desired[:, :len(keys)], keys),
            constraints=[SpacingConstraint(spacing),
                         XYBoundaryConstraint(boundary)],
            plot_comp=plot_comp,
            driver=driver,
            expected_cost=1.5)
    return _topfarm_obj


@pytest.mark.parametrize('driver,tol', [
    (EasyScipyOptimizeDriver(disp=False), 1e-4),
    (EasyScipyOptimizeDriver(tol=1e-3, disp=False), 1e-2),
    (EasyScipyOptimizeDriver(maxiter=14, disp=False), 1e-1),
    (EasyScipyOptimizeDriver(optimizer='COBYLA', tol=1e-3, disp=False), 1e-2),
    (EasySimpleGADriver(max_gen=10, pop_size=100, bits={'x': [12] * 3, 'y':[12] * 3}, random_state=1), 1e-1),
    (EasyPyOptSparseIPOPT(), 1e-4),
    (EasyPyOptSparseSNOPT(), 1e-4),
][:])
def test_optimizers(driver, tol, topfarm_generator_scalable):
    if driver.__class__.__name__ == "PyOptSparseMissingDriver":
        pytest.xfail("Driver missing")
    tf = topfarm_generator_scalable(driver)
    tf.evaluate()
    cost, _, recorder = tf.optimize({'x': [6., 6., 1.], 'y': [-.01, -8., 1.]})

    tb_pos = tf.turbine_positions[:, :2]

    assert sum((tb_pos[2] - tb_pos[0])**2) > 2**2 - tol  # check min spacing
    assert tb_pos[1][0] < 6 + tol  # check within border
    tf.plot_comp.show()
    if isinstance(driver, EasySimpleGADriver):
        assert cost == recorder['cost'].min()
    else:
        np.testing.assert_array_almost_equal(tb_pos, optimal[:, :2], -int(np.log10(tol)))


@pytest.mark.parametrize('driver,tol,N', [
    (EasyScipyOptimizeDriver(disp=False), 1e-4, 29),
    (EasyPyOptSparseSNOPT(), 1e-4, 39),
    # COBYLA no longer works with scaling.
    # See issue on Github: https://github.com/OpenMDAO/OpenMDAO/issues/942
    # It can therefore requires 120 iterations instead of 104
    #    (EasyScipyOptimizeDriver(optimizer='COBYLA', tol=1e-3, disp=False), 1e-2, 104),
    (EasyScipyOptimizeDriver(optimizer='COBYLA', tol=1e-3, disp=False), 1e-2, 120),
    (EasyPyOptSparseIPOPT(), 1e-4, 25),
][:])
@pytest.mark.parametrize('cost_scale,cost_offset', [(1, 0),
                                                    (1000, 0),
                                                    (1, 200),
                                                    (0.001, 200),
                                                    (0.001, -200),
                                                    (1000, 200),
                                                    (1000, -200)
                                                    ])
def test_optimizers_scaling(driver, tol, N, cost_scale, cost_offset, topfarm_generator_scalable):
    if driver.__class__.__name__ == "PyOptSparseMissingDriver":
        pytest.xfail("Driver missing")

    tf = topfarm_generator_scalable(driver, cost_scale=cost_scale, cost_offset=cost_offset)
    _, _, recorder = tf.optimize()
    uta.assertLessEqual(recorder.num_cases, N)

    tb_pos = tf.turbine_positions[:, :2]
    tf.plot_comp.show()

    assert sum((tb_pos[2] - tb_pos[0])**2) > 2**2 - tol  # check min spacing
    assert tb_pos[1][0] < 6 + tol  # check within border
    np.testing.assert_array_almost_equal(tb_pos, optimal[:, :2], -int(np.log10(tol)))


def find_optimal_scaling(topfarm_generator_scalable):
    i = 0.000001
    res = []
    while i < 100000:
        driver = EasyScipyOptimizeDriver(disp=False, tol=1e-6 * i)
        tf = topfarm_generator_scalable(driver)
        tf.model.get_objectives()['cost_comp.cost']['scaler'] = i

        cost, _, recorder = tf.optimize()
        N = recorder.num_cases
        res.append((i, N, cost))
        print(i, N, cost)
        i *= 2

    import matplotlib.pyplot as plt
    plt.figure()
    res = np.array(res)
    plt.plot(res[:, 0], res[:, 1])
    plt.xscale('log')
    plt.show()


@pytest.mark.parametrize('randomize_func', [RandomizeTurbinePosition_Circle(1),
                                            RandomizeTurbinePosition_Square(1),
                                            RandomizeTurbinePosition_Normal(1)])
def test_random_search_driver_position(topfarm_generator, randomize_func):
    np.random.seed(1)
    driver = EasyRandomSearchDriver(randomize_func, max_iter=1000)
    tf = topfarm_generator(driver, spacing=1)
    tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tol = 1e-1
    assert tb_pos[1][0] < 6 + tol  # check within border
    np.testing.assert_array_almost_equal(tb_pos, [[3, -3], [6, -7], [4, -3]], -int(np.log10(tol)))


def test_random_search_driver_type_and_position(topfarm_generator):
    np.random.seed(1)

    tf = TopFarmProblem(
        {'x': [1, 6, 6], 'y': [1, 0, -8], 'type': ([0, 0, 0], 0, 3)},
        cost_comp=DummyCost(desired, ['x', 'y', 'type']),
        constraints=[SpacingConstraint(1),
                     XYBoundaryConstraint(boundary)],
        driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbineTypeAndPosition(1), max_iter=2000),
    )
    _, state, _ = tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tol = 1e-1
    assert tb_pos[1][0] < 6 + tol  # check within border

    np.testing.assert_array_almost_equal(tb_pos, [[3, -3], [6, -7], [4, -3]], -int(np.log10(tol)))
    np.testing.assert_array_equal(state['type'], [1, 2, 3])


def test_random_search_driver_randomize_all_uniform():
    np.random.seed(1)

    class Cost():
        i = 0

        def __call__(self, *args, **kwargs):
            self.i += 1
            return self.i

    cost_comp = CostModelComponent(
        input_keys=['x', 'y', 'type'],
        n_wt=2,
        cost_function=Cost(),
        income_model=True)

    tf = TopFarmProblem(
        {'x': ([1, 6], [0, 1], [5, 6]), 'y': ([-1., 0], -6, 0), 'type': ([3, 3], 3, 8)},
        cost_comp=cost_comp,
        constraints=[],
        driver=EasyRandomSearchDriver(randomize_func=RandomizeAllUniform(['x', 'type']), max_iter=600, disp=False),
    )
    _, state, recorder = tf.optimize()

    # check that integer design variables are somewhat evenly distributed
    x, y, t = recorder['x'], recorder['y'], recorder['type']
    for arr, l, u in [(x[:, 0], 0, 5), (x[:, 1], 1, 6), (t[:, 0], 3, 8)]:
        count = [(arr == i).sum() for i in range(l, u + 1)]
        npt.assert_equal(601, sum(count))
        npt.assert_array_less(600 / len(count) * .70, count)

    count, _ = np.histogram(y[:, 0], np.arange(-6, 1))
    npt.assert_equal(y.shape[0], sum(count))
    npt.assert_array_less(600 / len(count) * .70, count)


def test_random_search_driver_RandomizeAllRelativeMaxStep(topfarm_generator):
    np.random.seed(1)

    tf = TopFarmProblem(
        {'x': [1, 6, 6], 'y': [1, 0, -8], 'type': ([0, 0, 0], 0, 3)},
        cost_comp=DummyCost(desired, ['x', 'y', 'type']),
        constraints=[SpacingConstraint(1),
                     XYBoundaryConstraint(boundary)],
        driver=EasyRandomSearchDriver(randomize_func=RandomizeAllRelativeMaxStep(.01), max_iter=2000),
    )
    _, state, _ = tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tol = 1e-1
    assert tb_pos[1][0] < 6 + tol  # check within border

    np.testing.assert_array_almost_equal(tb_pos, [[3, -3], [6, -7], [4, -3]], -int(np.log10(tol)))
    np.testing.assert_array_equal(np.round(state['type']), [1, 2, 3])



from topfarm.constraint_components.capacity import CapacityConstraint
import numpy as np
import topfarm
from topfarm.tests.test_files import xy3tb
from topfarm._topfarm import TopFarmProblem
from topfarm.easy_drivers import EasySimpleGADriver


def test_capacity_as_penalty():
    tf = xy3tb.get_tf(design_vars={topfarm.type_key: ([0, 0, 0], 0, 2)},
                      constraints=[CapacityConstraint(5, rated_power_array=[100, 10000, 10])],
                      driver=EasySimpleGADriver(),
                      plot_comp=None)

    # check normal result that satisfies the penalty
    assert tf.evaluate()[0] == 141.0
    # check penalized result if capacity constraint is not satisfied
    assert tf.evaluate({'type': np.array([0, 1, 1])})[0] == 1e10 + 15.1


def test_capacity_tf():
    # 15 turbines, 5 different types, 50MW max installed capacity
    n_wt = 15
    rated_power_array_kW = np.linspace(1, 10, int(n_wt / 3)) * 1e3

    inputtypes = np.tile(np.array([range(int(n_wt / 3))]), 3).flatten()
    tf = TopFarmProblem({'type': inputtypes},
                        constraints=[CapacityConstraint(max_capacity=50, rated_power_array=rated_power_array_kW)],
                        driver=EasySimpleGADriver()
                        )

    tf.evaluate()
    # case above the maximum allowed installed capacity, yes penalty
    assert tf["totalcapacity"] == 82.5
    assert tf['penalty_comp.penalty_capacity_comp_50'] == 32.5

    # set all turbines type 0, still 15 turbines and re-run the problem
    tf.evaluate({'type': inputtypes * 0})
    # case below the maximum allowed installed capacity, no penalty
    assert tf["totalcapacity"] == 15
    assert tf['penalty_comp.penalty_capacity_comp_50'][0] == 0.0



from openmdao.drivers.doe_generators import FullFactorialGenerator,\
    ListGenerator
from openmdao.drivers.doe_driver import DOEDriver
from topfarm.cost_models.dummy import DummyCost
import numpy as np
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.tests import npt
from topfarm import TopFarmProblem
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint


optimal = np.array([(0, 2, 4, 1), (4, 2, 1, 0)])


def test_2level_turbineType_and_XYZ_optimization():
    design_vars = {k: v for k, v in zip('xy', optimal.T)}
    design_vars['z'] = (optimal[:, 2], 0, 4)
    xyz_problem = TopFarmProblem(
        design_vars,
        cost_comp=DummyCost(optimal, ['x', 'y', 'z', 'type']),
        constraints=[SpacingConstraint(2), XYBoundaryConstraint([(0, 0), (4, 4)], 'square')],
        driver=EasyScipyOptimizeDriver(disp=False))
    tf = TopFarmProblem(
        {'type': ([0, 0], 0, 1)},
        cost_comp=xyz_problem,
        driver=DOEDriver(FullFactorialGenerator(2)))
    cost = tf.optimize()[0]
    assert cost == 0


def test_3level_type_multistart_XYZ_optimization():
    design_vars = {k: v for k, v in zip('xy', optimal.T)}
    design_vars['z'] = (optimal[:, 2], 0, 4)
    xyz_problem = TopFarmProblem(
        design_vars,
        cost_comp=DummyCost(optimal, ['x', 'y', 'z', 'type']),
        constraints=[SpacingConstraint(2), XYBoundaryConstraint([(0, 0), (4, 4)], 'square')],
        driver=EasyScipyOptimizeDriver(disp=False))

    initial_xyz_problem = TopFarmProblem(
        design_vars={k: v for k, v in zip('xyz', optimal.T)},
        cost_comp=xyz_problem,
        driver=DOEDriver(ListGenerator([[('x', [0, 4]), ('y', [2, 2]), ('z', [4, 1])]])))

    tf = TopFarmProblem(
        {'type': ([0, 0], 0, 1)},
        cost_comp=initial_xyz_problem,
        driver=DOEDriver(FullFactorialGenerator(2)))

    cost, _, recorder = tf.optimize()
    best_index = np.argmin(recorder.get('cost'))
    initial_xyz_recorder = recorder['recorder'][best_index]
    xyz_recorder = initial_xyz_recorder.get('recorder')[0]
    npt.assert_almost_equal(xyz_recorder['cost'][-1], cost)



from openmdao.drivers.scipy_optimizer import ScipyOptimizeDriver
# from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver  # version 2.5.0 has bug see Issue #874
from topfarm.drivers.genetic_algorithm_driver import SimpleGADriver
from topfarm.drivers.random_search_driver import RandomSearchDriver
import sys
import numpy as np
import openmdao


class EasyDriverBase():
    expected_cost = 1

    def get_desvar_kwargs(self, model, desvar_name, desvar_values):
        if len(desvar_values) == 4:
            kwargs = {'lower': desvar_values[1], 'upper': desvar_values[2]}
        else:
            kwargs = {}
        return kwargs

    @property
    def supports_expected_cost(self):
        return True


class EasyScipyOptimizeDriver(ScipyOptimizeDriver, EasyDriverBase):

    def __init__(self, optimizer='SLSQP', maxiter=200, tol=1e-6, disp=True, **kwargs):
        """
        Parameters
        ----------
        optimizer : {'COBYLA', 'SLSQP'}
            Gradients are only supported by SLSQP
        maxiter : int
            Maximum number of iterations.
        tol : float
            Tolerance for termination. For detailed control, use solver-specific options.
        disp : bool
            Set to False to prevent printing of Scipy convergence messages
        """
        ScipyOptimizeDriver.__init__(self)
        self.options.update({'optimizer': optimizer, 'maxiter': maxiter, 'tol': tol, 'disp': disp})
        if kwargs:
            self.options.update(kwargs)

    def get_desvar_kwargs(self, model, desvar_name, desvar_values):
        kwargs = super().get_desvar_kwargs(model, desvar_name, desvar_values)
        if self.options['optimizer'] == 'SLSQP':
            # Upper and lower disturbs SLSQP when running with constraints. Add limits as constraints
            model.add_constraint(desvar_name, kwargs.get('lower', None), kwargs.get('upper', None))
            kwargs = {'lower': np.nan, 'upper': np.nan}  # Default +/- sys.float_info.max does not work for SLSQP
            ref0 = 0
            ref1 = 1
            # TODO: Check if the following improves performance
            # if len(desvar_values) == 4:
            #     ref0 = np.min(desvar_values[1])
            #     ref1 = np.max(desvar_values[2])

            kwargs = {'ref0': ref0, 'ref': ref1, 'lower': np.nan, 'upper': np.nan}
        elif openmdao.__version__ != '2.6.0' and self.options['optimizer'] == 'COBYLA':
            # COBYLA does not work with ref-setting in openmdao 2.6.0.
            # See issue on Github: https://github.com/OpenMDAO/OpenMDAO/issues/942
            if len(desvar_values) == 4:
                ref0 = np.min(desvar_values[1])
                ref1 = np.max(desvar_values[2])
                l, u = [lu * (ref1 - ref0) + ref0 for lu in [desvar_values[1], desvar_values[2]]]
                kwargs = {'ref0': ref0, 'ref': ref1, 'lower': l, 'upper': u}
        return kwargs

    @property
    def supports_expected_cost(self):
        return not (openmdao.__version__ == '2.6.0' and self.options['optimizer'] == 'COBYLA')


try:
    class PyOptSparseMissingDriver(object):
        options = {}

    from openmdao.drivers.pyoptsparse_driver import pyOptSparseDriver
#     Not working:
#     capi_return is NULL
#     Call-back cb_slfunc_in_slsqp__user__routines failed.
#
#     class EasyPyOptSparseSLSQP(pyOptSparseDriver):
#         def __init__(self, maxit=200, acc=1e-6):
#             pyOptSparseDriver.__init__(self)
#             self.options.update({'optimizer': 'SLSQP'})
#             self.opt_settings.update({'MAXIT': maxit, 'ACC': acc})

    class EasyPyOptSparseIPOPT(pyOptSparseDriver):
        def __init__(self, max_iter=200):
            pyOptSparseDriver.__init__(self)
            self.options.update({'optimizer': 'IPOPT'})
            self.opt_settings.update({'linear_solver': 'ma27', 'max_iter': max_iter,
                                      'start_with_resto': 'yes',
                                      'expect_infeasible_problem': 'yes'})

    from pyoptsparse.pyIPOPT.pyIPOPT import pyipoptcore
    if pyipoptcore is None:
        setattr(sys.modules[__name__], 'EasyPyOptSparseIPOPT', PyOptSparseMissingDriver)

    class EasyPyOptSparseSNOPT(pyOptSparseDriver, EasyDriverBase):
        def __init__(self, major_iteration_limit=200, major_feasibility_tolerance=1e-6, major_optimality_tolerance=1e-6,
                     difference_interval=1e-6, function_precision=1e-8,
                     Print_file='SNOPT_print.out', Summary_file='SNOPT_summary.out', print_results=False):
            """For information about the arguments see
            https://web.stanford.edu/group/SOL/software/snoptHelp/whgdata/whlstt9.htm#9
            """
            pyOptSparseDriver.__init__(self)
            self.options.update({'optimizer': 'SNOPT', 'print_results': print_results})
            self.opt_settings.update({
                'Major feasibility tolerance': major_feasibility_tolerance,
                'Major optimality tolerance': major_optimality_tolerance,
                'Difference interval': difference_interval,
                'Hessian full memory': None,
                'Function precision': function_precision,
                'Major iterations limit': major_iteration_limit,
                'Print file': Print_file,
                'Summary file': Summary_file,
                'Major step limit': 2.0})

        def get_desvar_kwargs(self, model, desvar_name, desvar_values):
            kwargs = EasyDriverBase.get_desvar_kwargs(self, model, desvar_name, desvar_values)
            if len(desvar_values) == 4:
                ref0 = np.min(desvar_values[1])
                ref1 = np.max(desvar_values[2])
                l, u = desvar_values[1], desvar_values[2]
                kwargs = {'ref0': ref0, 'ref': ref1, 'lower': l, 'upper': u}
            return kwargs


except ModuleNotFoundError:
    for n in ['EasyPyOptSparseSLSQP', 'EasyPyOptSparseIPOPT', 'EasyPyOptSparseSNOPT']:
        setattr(sys.modules[__name__], n, PyOptSparseMissingDriver)


class EasySimpleGADriver(SimpleGADriver, EasyDriverBase):
    def __init__(self, max_gen=100, pop_size=25, Pm=None, Pc=.5, elitism=True, bits={}, debug_print=[], run_parallel=False, random_state=None):
        """SimpleGA driver with optional arguments

        Parameters
        ----------
        max_gen : int
            Number of generations before termination.
        pop_size : int
            Number of points in the GA.
        pm : float
            Probability of mutation.
        pc : float
             Probability of cross over.
        elitism : bool, optional
            If True, replace worst performing point with best from previous generation each iteration.
        bits : dict, optional
            Number of bits of resolution. Default is an empty dict, where every
            unspecified variable is assumed to be integer, and the number of
            bits is calculated automatically.
            If you have a continuous var, you should set a bits value as a key
            in this dictionary.
            Ex: {'x':16,'y':16}
        debug_print : list, optional
            List of what type of Driver variables to print at each iteration. Valid items in list are ‘desvars’,’ln_cons’,’nl_cons’,’objs’
        run_parallel : bool
            Set to True to execute the points in a generation in parallel.
        """
        SimpleGADriver.__init__(self, max_gen=max_gen, pop_size=pop_size, Pm=Pm, Pc=Pc, elitism=elitism,
                                bits=bits, debug_print=debug_print, run_parallel=run_parallel)
        self.supports['inequality_constraints'] = False
        self.supports['equality_constraints'] = False
        if random_state is not None:
            self._randomstate = random_state


class EasyRandomSearchDriver(RandomSearchDriver, EasyDriverBase):
    def __init__(self, randomize_func, max_iter=100, max_time=600, disp=False, run_parallel=False):
        """Easy initialization of RandomSearchDriver

        Parameters
        ----------
        randomize_func : f(desvar_dict)
            Function to randomize desired variables of desvar_dict
        max_iter : int, optional
            Maximum iterations
        max_time : int, optional
            Maximum time in seconds
        disp : bool
        """
        RandomSearchDriver.__init__(self, randomize_func=randomize_func,
                                    max_iter=max_iter, max_time=max_time, disp=disp, run_parallel=run_parallel)



from openmdao.core.explicitcomponent import ExplicitComponent
from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
import numpy as np


class PenaltyComponent(ExplicitComponent):
    def __init__(self, constraint_comps, as_penalty):
        super().__init__()
        self.constraint_comps = constraint_comps
        self.as_penalty = as_penalty

    def setup(self):
        for comp in self.constraint_comps:
            self.add_input('penalty_' + comp.const_id, val=0.0)
        self.add_output('penalty', val=0.0)

    def compute(self, inputs, outputs):
        penalties = np.array([inputs[x] for x in inputs])
        if self.as_penalty:
            penalty = sum(penalties)
        else:
            penalty = 0.0
        outputs['penalty'] = penalty


class PostPenaltyComponent(ExplicitComponent):
    def __init__(self, constraint_comps, as_penalty):
        super().__init__()
        self.constraint_comps = constraint_comps
        self.as_penalty = as_penalty

    def setup(self):
        for comp in self.constraint_comps:
            self.add_input('post_penalty_' + comp[0], val=0.0)
        self.add_output('post_penalty', val=0.0)

    def compute(self, inputs, outputs):
        penalties = np.array([inputs[x] for x in inputs])
        if self.as_penalty:
            penalty = sum(penalties)
        else:
            penalty = 0.0
        outputs['post_penalty'] = penalty



'''
Created on 17. jul. 2018

@author: mmpe
'''
from topfarm.recorders import TopFarmListRecorder, split_record_id,\
    recordid2filename
import numpy as np
from openmdao.drivers.doe_generators import ListGenerator
import pytest
from topfarm.cost_models.dummy import DummyCost
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from openmdao.drivers.doe_driver import DOEDriver
from numpy import testing as npt
from topfarm.tests.test_files import tfp
import os
from topfarm.tests.test_fuga import test_pyfuga
from topfarm.plotting import PlotComp, NoPlot, XYPlotComp
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm._topfarm import TopFarmProblem
from topfarm.tests.test_fuga.test_pyfuga import get_fuga


@pytest.fixture
def tf_generator():
    def tf(xy_boundary=[(0, 0), (4, 4)], z_boundary=(0, 4), xy_boundary_type='square', **kwargs):
        optimal = [(0, 2, 4), (4, 2, 1)]
        xyz = np.array([(0, 1, 0), (1, 1, 1)])
        p1 = DummyCost(optimal, 'xyz')
        design_vars = dict(zip('xy', xyz.T))
        design_vars['z'] = (xyz[:, 2], z_boundary[0], z_boundary[1])
        k = {'design_vars': design_vars,
             'cost_comp': p1,
             'driver': EasyScipyOptimizeDriver(optimizer='COBYLA', disp=False, maxiter=10),
             }
        k.update(kwargs)

        return TopFarmProblem(
            constraints=[XYBoundaryConstraint(xy_boundary, xy_boundary_type),
                         SpacingConstraint(2)],
            **k
        )

    return tf


def test_ListRecorder():
    from openmdao.api import Problem, IndepVarComp
    from openmdao.test_suite.components.paraboloid import Paraboloid

    prob = Problem()
    model = prob.model

    model.add_subsystem('p1', IndepVarComp('x', 0.), promotes=['*'])
    model.add_subsystem('p2', IndepVarComp('y', 0.), promotes=['*'])
    model.add_subsystem('comp', Paraboloid(), promotes=['*'])

    model.add_design_var('x', lower=-10, upper=10)
    model.add_design_var('y', lower=-10, upper=10)
    model.add_objective('f_xy')

    xyf = [[0.98, 4.30, 74.1844],
           [2.06, 0.90, 23.7476],
           [-1.53, 2.92, 60.9397],
           [-1.25, 7.84, 145.4481]]
    prob.driver = DOEDriver(ListGenerator([[('x', xy[0]), ('y', xy[1])] for xy in xyf]))
    recorder = TopFarmListRecorder()
    recorder._initialize_database()
    recorder._cleanup_abs2meta()
    recorder.record_iteration_problem(None, None, None)
    recorder.record_iteration_system(None, None, None)
    recorder.record_iteration_solver(None, None, None)
    recorder.record_viewer_data(None)
    recorder.record_metadata_solver(None)
    recorder.record_derivatives_driver(None, None, None)
    recorder.shutdown()

    prob.driver.add_recorder(recorder)
    prob.driver.recording_options['record_desvars'] = True
    prob.driver.recording_options['includes'] = ['*']
    prob.driver.recording_options['record_inputs'] = True

    prob.setup()
    prob.run_driver()
    prob.cleanup()

    assert recorder.num_cases == 4

    npt.assert_array_equal(recorder.get('counter'), range(1, 5))
    npt.assert_array_equal(recorder['counter'], range(1, 5))

    npt.assert_array_almost_equal(recorder.get(['x', 'y', 'f_xy']), xyf, 4)
    for xyf, k in zip(xyf[0], ['x', 'y', 'f_xy']):
        npt.assert_allclose(recorder[k][0], xyf)

    with pytest.raises(KeyError, match="missing"):
        recorder.get('missing')


def test_TopFarmListRecorderAnimation(tf_generator):
    try:
        from matplotlib import animation
        animation.writers['ffmpeg']
    except Exception:
        pytest.xfail("No matplotlib, animation or ffmpeg writer")

    tf = tf_generator()
    _, _, recorder = tf.optimize()
    # # Generate test file:
#    recorder.save('topfarm/tests/test_files/recordings/COBYLA_10iter.pkl')
    fn = tfp + "/tmp/test.mp4"
    if os.path.exists(fn):
        os.remove(fn)
    recorder.animate_turbineXY(duration=5, filename=fn)
    assert os.path.isfile(fn)


def test_NestedTopFarmListRecorder(tf_generator):
    optimal = [(0, 2, 4, 1), (4, 2, 1, 0)]
    type_lst = [[0, 0],
                [1, 0],
                [0, 1]]
    p1 = DummyCost(optimal_state=optimal,
                   inputs=['x', 'y', 'z', 'type'])
    p2 = tf_generator(cost_comp=p1,
                      driver=EasyScipyOptimizeDriver(disp=False))

    tf = TopFarmProblem(
        {'type': ([0, 0], 0, 1)},
        cost_comp=p2,
        driver=DOEDriver(ListGenerator([[('type', t)] for t in type_lst])))

    cost, _, recorder = tf.optimize()
    npt.assert_almost_equal(cost, 0)
    npt.assert_array_almost_equal(recorder.get('type'), type_lst)
    npt.assert_array_almost_equal(recorder.get('cost'), [1, 0, 2])

    for sub_rec in recorder.get('recorder'):
        npt.assert_array_almost_equal(np.array([sub_rec[k][-1] for k in ['x', 'y', 'z']]).T, np.array(optimal)[:, :3])


@pytest.mark.parametrize('record_id,filename,load_case', [
    ("test", 'recordings/test.pkl', 'latest'),
    ("test.pkl", 'recordings/test.pkl', 'latest'),
    ("recordings/test.pkl", 'recordings/test.pkl', 'latest'),
    ("topfarm/tests/test_files/tmp/test", 'topfarm/tests/test_files/tmp/test.pkl', 'latest'),
    ("topfarm/tests/test_files/tmp/test.pkl", 'topfarm/tests/test_files/tmp/test.pkl', 'latest'),
    ("c:/tmp/test", 'c:/tmp/test.pkl', 'latest'),
    ("c:/tmp/test:latest", 'c:/tmp/test.pkl', 'latest'),
    ("test:", 'recordings/test.pkl', 'latest'),
    ("test:latest", 'recordings/test.pkl', 'latest'),
    ("test:best", 'recordings/test.pkl', 'best'),
    ("test:5", 'recordings/test.pkl', '5'),
    ("test:none", 'recordings/test.pkl', 'none'),
    (None, "", "")
])
def test_recordid2filename(record_id, filename, load_case):
    fn, lc = TopFarmListRecorder().recordid2filename(record_id)
    assert fn == filename
    assert lc == load_case
    fn, lc = recordid2filename(record_id)
    assert fn == filename
    assert lc == load_case


@pytest.mark.parametrize('record_id,record_name,load_case', [
    ("test", 'test', 'latest'),
    ("test.pkl", 'test.pkl', 'latest'),
    ("recordings/test.pkl", 'recordings/test.pkl', 'latest'),
    ("topfarm/tests/test_files/tmp/test", 'topfarm/tests/test_files/tmp/test', 'latest'),
    ("topfarm/tests/test_files/tmp/test.pkl", 'topfarm/tests/test_files/tmp/test.pkl', 'latest'),
    ("c:/tmp/test", 'c:/tmp/test', 'latest'),
    ("c:/tmp/test:latest", 'c:/tmp/test', 'latest'),
    ("test:", 'test', 'latest'),
    ("test:latest", 'test', 'latest'),
    ("test:best", 'test', 'best'),
    ("test:none", 'test', 'none'),
    (None, "", "")
])
def test_split_recordid(record_id, record_name, load_case):
    n, lc = TopFarmListRecorder().split_record_id(record_id)
    assert n == record_name
    assert lc == load_case
    n, lc = split_record_id(record_id)
    assert n == record_name
    assert lc == load_case


@pytest.mark.parametrize('load_case,n,cost',
                         [("latest", 11, 2.6187984),
                          ('best', 11, 2.6187984),
                          ("3", 3, 28)])
def test_TopFarmListRecorderLoad(load_case, n, cost):
    fn = tfp + 'recordings/COBYLA_10iter:%s' % load_case
    rec = TopFarmListRecorder().load(fn)
    npt.assert_equal(rec.num_cases, n)
    npt.assert_almost_equal(rec.get('cost')[-1], cost)


@pytest.mark.parametrize('load_case', [("none"),
                                       ('0')])
def test_TopFarmListRecorderLoad_none(load_case):
    # load case is "none"
    fn = tfp + 'recordings/COBYLA_10iter:%s' % load_case
    rec = TopFarmListRecorder().load(fn)
    assert rec.num_cases == 0


@pytest.mark.parametrize('fn', [(None),
                                ('MissingFile')])
def test_TopFarmListRecorderLoad_Nothing(fn):
    # No such file
    with pytest.raises(FileNotFoundError, match=r"No such file '.*'"):
        TopFarmListRecorder().load(fn)
    rec = TopFarmListRecorder().load_if_exists(fn)
    assert rec.num_cases == 0


@pytest.mark.parametrize('load_case,n_rec,n_fev', [('', 55, 1),
                                                   ('none', 53, 52),
                                                   (40, 207, 166)])
def test_TopFarmListRecorder_continue(tf_generator, load_case, n_rec, n_fev, get_fuga):

    D = 80.0
    D2 = 2 * D + 10
    init_pos = np.array([(0, 2 * D), (0, 0), (0, -2 * D)])
    init_pos[:, 0] += [-40, 0, 40]

    pyFuga = get_fuga(init_pos[:, 0], init_pos[:, 1], wind_atlas='MyFarm/north_pm45_only.lib')
    boundary = [(-D2, -D2), (D2, D2)]
    plot_comp = XYPlotComp()
    plot_comp = NoPlot()
    tf = TopFarmProblem(
        dict(zip('xy', init_pos.T)),
        cost_comp=pyFuga.get_TopFarm_cost_component(),
        constraints=[SpacingConstraint(2 * D),
                     XYBoundaryConstraint(boundary, 'square')],
        driver=EasyScipyOptimizeDriver(tol=1e-10, disp=False),
        plot_comp=plot_comp,
        record_id=tfp + 'recordings/test_TopFarmListRecorder_continue:%s' % load_case,
        expected_cost=25)

    _, _, recorder = tf.optimize()
    # Create test file:
    # 1) delete file "test_files/recordings/test_TopFarmListRecorder_continue"
    # 2) Uncomment line below, run and recomment
#    if load_case == "": recorder.save()  # create test file
    npt.assert_equal(recorder.num_cases, n_rec)
    npt.assert_equal(tf.driver.result['nfev'], n_fev)

    tf.plot_comp.show()


def test_TopFarmListRecorder_continue_wrong_recorder(tf_generator):

    tf = TopFarmProblem(
        {'type': ([0, 0, 0], 0, 1)},
        cost_comp=DummyCost(np.array([[0, 1, 0]]).T, ['type']),
        driver=EasyScipyOptimizeDriver(disp=False),
        record_id=tfp + 'recordings/test_TopFarmListRecorder_continue:latest'
    )

    tf.optimize()
    assert 'type' in tf.recorder.keys()
    assert 'x' not in tf.recorder.keys()


@pytest.mark.parametrize('rec_id,sn,fn', [(tfp + "recordings/tmp", None, tfp + "recordings/tmp"),
                                          (None, tfp + "recordings/tmp2", tfp + "recordings/tmp2")])
def test_TopFarmListRecorder_save(tf_generator, rec_id, sn, fn):
    def remove_file():
        if os.path.isfile(fn + ".pkl"):
            os.remove(fn + ".pkl")
    remove_file()
    tf = tf_generator(record_id=rec_id)
    _, _, recorder = tf.optimize()
    recorder.save(sn)
    npt.assert_array_equal(recorder.get('cost'), TopFarmListRecorder().load(fn).get('cost'))
    remove_file()



from openmdao.core.driver import Driver, RecordingDebugging
from six import iteritems
import numpy as np
from openmdao.core.analysis_error import AnalysisError
import topfarm
import time
from openmdao.utils.concurrent import concurrent_eval
from openmdao.utils.record_util import create_local_meta


class RandomSearchDriver(Driver):

    def __init__(self, randomize_func, **kwargs):
        self.randomize_func = randomize_func
        super().__init__(**kwargs)

        # What we support
        self.supports['integer_design_vars'] = True

        # What we don't support yet
        self.supports['inequality_constraints'] = False
        self.supports['equality_constraints'] = False
        self.supports['multiple_objectives'] = False
        self.supports['two_sided_constraints'] = False
        self.supports['linear_constraints'] = False
        self.supports['simultaneous_derivatives'] = False
        self.supports['active_set'] = False

        self._desvar_idx = {}

        # random state can be set for predictability during testing
        self._randomstate = None

    def _declare_options(self):
        """
        Declare options before kwargs are processed in the init method.
        """
        self.options.declare('max_iter', 200,
                             desc='Maximum number of iterations (set to None to disable)')
        self.options.declare('max_time', 600,
                             desc='Maximum time in seconds (set to None to disable)')
        self.options.declare('disp', True,
                             desc='Set to False to prevent printing')
        self.options.declare('run_parallel', False,
                             desc='Flag to run in parallel or not, requires mpi')

    def _setup_driver(self, problem):
        """
        Prepare the driver for execution.

        This is the final thing to run during setup.

        Parameters
        ----------
        problem : <Problem>
            Pointer to the containing problem.
        """
        super()._setup_driver(problem)

        model_mpi = None
        comm = self._problem.comm
        if not self.options['run_parallel']:
            comm = None

        if len(self._objs) > 1:
            msg = 'RandomSearchDriver currently does not support multiple objectives.'
            raise RuntimeError(msg)

        if len(self._cons) > 0:
            msg = 'RandomSearchDriver currently does not support constraints.'
            raise RuntimeError(msg)

        self.comm = comm

    def run(self):
        """
        Excute the genetic algorithm.

        Returns
        -------
        boolean
            Failure flag; True if failed to converge, False is successful.
        """
        model = self._problem.model

        # Size design variables.
        desvars = self._designvars
        count = 0
        for name, meta in iteritems(desvars):
            size = meta['size']
            self._desvar_idx[name] = (count, count + size)
            count += size

        lower_bound = np.empty((count, ))
        upper_bound = np.empty((count, ))

        x0 = np.empty(count)
        desvar_vals = self.get_design_var_values()

        # Figure out bounds vectors and initial design vars
        for name, meta in iteritems(desvars):
            i, j = self._desvar_idx[name]
            lower_bound[i:j] = meta['lower']
            upper_bound[i:j] = meta['upper']
            x0[i:j] = desvar_vals[name]
        max_iter = self.options['max_iter'] or 1e20
        max_time = self.options['max_time'] or 1e20
        assert max_iter < 1e20 or max_time < 1e20, "max_iter or max_time must be set"

        disp = self.options['disp']

        abs2prom = model._var_abs2prom['output']

        # Initial Design Vars
        desvar_vals = self.get_design_var_values()
        i = 0
        for name, meta in iteritems(self._designvars):
            size = meta['size']
            x0[i:i + size] = desvar_vals[name]
            i += size

        obj_value_x0, success = self.objective_callback(x0, record=True)
        x1 = x0.copy()
        n_iter = 0

        desvar_info = [(abs2prom[name], *self._desvar_idx[name], lower_bound, upper_bound)
                       for name, meta in iteritems(desvars)]
        desvar_dict = {name: (x0[i:j].copy(), lbound[i:j], ubound[i:j]) for (name, i, j, lbound, ubound) in desvar_info}
        start = time.time()

        comm = self.comm

        while n_iter < max_iter and time.time() - start < max_time:

            for name, i, j, _, _ in desvar_info:
                desvar_dict[name][0][:] = x0[i:j].copy()

            if comm is not None:
                # We do it in parallel: One case per CPU available
                cases = []
                if comm.rank == 0:
                    for ii in range(comm.size):
                        desvar_dict = self.randomize_func(desvar_dict)
                        x1 = x0.copy()
                        for name, i, j, _, _ in desvar_info:
                            x1[i:j] = desvar_dict[name][0][:]
                        cases.append(((x1, ii), None))

                # Let's make sure we have the same cases everywhere
                cases = comm.bcast(cases, root=0)

                results = concurrent_eval(self.objective_callback, cases, comm, allgather=True)
                one_success = False
                for i, result in enumerate(results):
                    returns, traceback = result
                    obj_value_x1, success = returns
                    if success and obj_value_x1 < obj_value_x0:
                        one_success = True
                        x0 = cases[i][0][0].copy()
                        obj_value_x0 = obj_value_x1
                        # n_iter += 1
                    elif obj_value_x1 < 1e10:
                        one_success = True
                    # n_iter += 1
                if one_success:
                    n_iter += 1
                    obj_value_x0, success = self.objective_callback(x0, record=True)
                    if disp and comm.rank == 0:
                        # obj_value_x0, success = self.objective_callback(x0, record=False)
                        print('rank:', comm.rank, n_iter, obj_value_x0)
            else:
                # We only use one CPU

                desvar_dict = self.randomize_func(desvar_dict)
                for name, i, j, _, _ in desvar_info:
                    x1[i:j] = desvar_dict[name][0][:]

                obj_value_x1, success = self.objective_callback(x1, record=False)
                if success and obj_value_x1 < obj_value_x0:
                    obj_value_x1, success = self.objective_callback(x1, record=True)
                    x0 = x1.copy()
                    obj_value_x0 = obj_value_x1
                    n_iter += 1
                    if disp:
                        print(n_iter, obj_value_x1)
                else:
                    if obj_value_x1 < 1e10:
                        n_iter += 1

                if not success or obj_value_x1 > obj_value_x0:
                    obj_value_x1, success = self.objective_callback(x0, record=True)
        return False

    def objective_callback(self, x, record=False):
        """
        Evaluate problem objective at the requested point.

        Parameters
        ----------
        x : ndarray
            Value of design variables.
        icase : int
            Case number, used for identification when run in parallel.

        Returns
        -------
        float
            Objective value
        bool
            Success flag, True if successful
        int
            Case number, used for identification when run in parallel.
        """
        model = self._problem.model
        success = 1

        for name in self._designvars:
            i, j = self._desvar_idx[name]
            self.set_design_var(name, x[i:j])

        # Execute the model
        if record:
            with RecordingDebugging('RandomSearch', self.iter_count, self) as rec:
                self.iter_count += 1
                try:
                    model._solve_nonlinear()

                # Tell the optimizer that this is a bad point.
                except AnalysisError:
                    model._clear_iprint()
                    success = 0

                for name, val in iteritems(self.get_objective_values()):
                    obj = val
                    break

                # Record after getting obj to assure they have
                # been gathered in MPI.
                rec.abs = 0.0
                rec.rel = 0.0
        else:
            try:
                model._solve_nonlinear()

            # Tell the optimizer that this is a bad point.
            except AnalysisError:
                model._clear_iprint()
                success = 0

            for name, val in iteritems(self.get_objective_values()):
                obj = val
                break
        return obj, success


def randomize_turbine_type(desvar_dict, i_wt=None):
    types, lbound, ubound = desvar_dict[topfarm.type_key]
    i_wt = i_wt or np.random.randint(len(types))
    types[i_wt] = np.random.randint(lbound[i_wt], ubound[i_wt] + 1)
    return desvar_dict


def xy_step_circle(max_step_xy):
    step = np.random.rand() * max(max_step_xy)
    theta = np.random.rand() * np.pi * 2
    return step * np.cos(theta), step * np.sin(theta)


def xy_step_square(max_step_xy):
    return (np.random.rand() * 2 - 1) * max_step_xy[0], (np.random.rand() * 2 - 1) * max_step_xy[1]


def xy_step_normal(max_step_xy):
    return np.random.normal(0, max_step_xy[0] / 2), np.random.normal(0, max_step_xy[1] / 2)


class RandomizeTurbinePosition():
    def __init__(self, max_step=None, xy_step_function=xy_step_circle):
        self.max_step = max_step
        self.xy_step_function = xy_step_function

    def __call__(self, desvar_dict, i_wt=None):
        i_wt = i_wt or np.random.randint(len(desvar_dict[topfarm.x_key][0]))

        max_step_xy = [self.max_step or (desvar_dict[xy][2][i_wt] - desvar_dict[xy][1][i_wt])
                       for xy in [topfarm.x_key, topfarm.y_key]]
        dxy = self.xy_step_function(max_step_xy)
        for (xy, lbound, ubound), dxy_ in zip([desvar_dict[topfarm.x_key], desvar_dict[topfarm.y_key]],
                                              dxy):
            xy[i_wt] = np.maximum(np.minimum(xy[i_wt] + dxy_, ubound[i_wt]), lbound[i_wt])
        return desvar_dict


class RandomizeTurbinePosition_Circle(RandomizeTurbinePosition):
    def __init__(self, max_step=None):
        RandomizeTurbinePosition.__init__(self, max_step=max_step, xy_step_function=xy_step_circle)


class RandomizeTurbinePosition_Square(RandomizeTurbinePosition):
    def __init__(self, max_step=None):
        RandomizeTurbinePosition.__init__(self, max_step=max_step, xy_step_function=xy_step_square)


class RandomizeTurbinePosition_Normal(RandomizeTurbinePosition):
    def __init__(self, max_step=None):
        RandomizeTurbinePosition.__init__(self, max_step=max_step, xy_step_function=xy_step_normal)


class RandomizeTurbineTypeAndPosition(RandomizeTurbinePosition):
    def __init__(self, max_step=None, xy_step_function=xy_step_circle):
        RandomizeTurbinePosition.__init__(self, xy_step_function=xy_step_function, max_step=max_step)

    def __call__(self, desvar_dict, i_wt=None):
        i_wt = i_wt or np.random.randint(len(desvar_dict[topfarm.x_key][0]))
        desvar_dict = randomize_turbine_type(desvar_dict, i_wt=i_wt)
        desvar_dict = RandomizeTurbinePosition.__call__(self, desvar_dict, i_wt=i_wt)
        return desvar_dict


class RandomizeAllUniform():
    def __init__(self, int_design_vars_lst=[]):
        """Initialize RandomizeAllUniform

        Parameters
        ----------
        int_design_vars_lst : list
            names of all integer design variables
        """
        self.int_design_vars_lst = int_design_vars_lst

    def __call__(self, desvar_dict):
        for name, (values, lbound, ubound) in desvar_dict.items():
            if name in self.int_design_vars_lst:
                if np.all(lbound == lbound[0]) and np.all(ubound == ubound[0]):
                    values[:] = np.random.randint(lbound[0], ubound[0] + 1, values.shape)
                else:
                    lbound_arr = np.zeros_like(values) + lbound
                    ubound_arr = np.zeros_like(values) + ubound
                    for i, (l, u) in enumerate(zip(lbound_arr, ubound_arr)):
                        values[i] = np.random.randint(l, u + 1)
            else:
                values[:] = np.random.uniform(lbound, ubound, values.shape)
        return desvar_dict


class RandomizeAllRelativeMaxStep():
    def __init__(self, rel_max_step=.1):
        self.rel_max_step = rel_max_step

    def __call__(self, desvar_dict):
        for _, (values, lbound, ubound) in desvar_dict.items():
            if values.dtype.kind == 'i':
                raise NotImplementedError()
            else:
                step = (np.random.random(values.shape) * 2 - 1) * (ubound - lbound) * self.rel_max_step
                values[:] = np.maximum(np.minimum(values + step, ubound), lbound)
        return desvar_dict


class RandomizeNUniform():
    def __init__(self, N, int_design_vars_lst=[]):
        """Initialize RandomizeNUniform

        Parameters
        ----------
        N : int
            Number of elements of the design vectors to randomize
        int_design_vars_lst : list
            names of all integer design variables
        """
        self.N = N
        self.int_design_vars_lst = int_design_vars_lst

    def __call__(self, desvar_dict):

        indexes = np.random.randint(0, len(list(desvar_dict.items())[0][1][0]), self.N)
        for name, (values, lbound, ubound) in desvar_dict.items():
            if name in self.int_design_vars_lst:
                if np.all(lbound == lbound[0]) and np.all(ubound == ubound[0]):
                    values[indexes] = np.random.randint(lbound[0], ubound[0] + 1, len(indexes))
                else:
                    lbound_arr = np.zeros_like(values) + lbound
                    ubound_arr = np.zeros_like(values) + ubound
                    for i in indexes:
                        values[i] = np.random.randint(lbound_arr[i], ubound_arr[i] + 1)
            else:
                values[indexes] = np.random.uniform(lbound[indexes], ubound[indexes], len(indexes))
        return desvar_dict



import numpy as np
from topfarm.cost_models.cost_model_wrappers import CostModelComponent,\
    AEPCostModelComponent
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm import TopFarmProblem
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.plotting import NoPlot, XYPlotComp
from topfarm.tests import npt

boundary = [(0, 0), (6, 0), (6, -10), (0, -10)]  # turbine boundaries
initial = np.array([[6, 0], [6, -8], [1, 1], [-1, -8]])  # initial turbine layouts
optimal_with_constraints = np.array([[2.5, -3], [6, -7], [4.5, -3], [3, -7]])  # optimal turbine layout
min_spacing = 2  # min distance between turbines
optimal = np.array([[3, -3], [7, -7], [4, -3], [3, -7]])  # desired turbine layouts


def get_tf(cost_comp, plot_comp=NoPlot()):
    return TopFarmProblem(
        dict(zip('xy', initial.T)),
        cost_comp=cost_comp,
        plot_comp=plot_comp,
        constraints=[SpacingConstraint(min_spacing), XYBoundaryConstraint(boundary)],
        driver=EasyScipyOptimizeDriver(disp=False))


def cost(x, y):
    opt_x, opt_y = optimal.T
    return np.sum((x - opt_x)**2 + (y - opt_y)**2)


def aep_cost(x, y):
    opt_x, opt_y = optimal.T
    return -np.sum((x - opt_x)**2 + (y - opt_y)**2)


def gradients(x, y):
    return (2 * x - 2 * optimal[:, 0]), (2 * y - 2 * optimal[:, 1])


def aep_gradients(x, y):
    return -(2 * x - 2 * optimal[:, 0]), -(2 * y - 2 * optimal[:, 1])


def test_CostModelComponent():
    tf = get_tf(CostModelComponent(['x', 'y'], 4, cost, gradients),)
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 5)


def testCostModelComponent_no_gradients():
    tf = get_tf(CostModelComponent(['x', 'y'], 4, cost, None))
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 4)


def testAEPCostModelComponent():
    tf = get_tf(AEPCostModelComponent(['x', 'y'], 4, aep_cost, aep_gradients))
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 5)


def test_maxiter_CostModelComponent():
    tf = get_tf(AEPCostModelComponent(['x', 'y'], 4, aep_cost, aep_gradients, max_eval=10))
    cost, state, recorder = tf.optimize()
    assert 10 <= tf.cost_comp.counter <= 11, tf.cost_comp.counter
    npt.assert_array_equal(recorder['AEP'][tf.cost_comp.n_func_eval], recorder['AEP'][tf.cost_comp.n_func_eval:])


def testCostModelComponentDiffShapeInput():
    def aep_cost(x, y, h):
        opt_x, opt_y = optimal.T
        return -np.sum((x - opt_x)**2 + (y - opt_y)**2) + h, {'add_out': sum(x)}
    cost_comp = AEPCostModelComponent(['x', 'y', ('h', 0)], 4, aep_cost, additional_output=[('add_out', 0)])
    tf = TopFarmProblem(
        dict(zip('xy', initial.T)),
        cost_comp=cost_comp,
        constraints=[SpacingConstraint(min_spacing), XYBoundaryConstraint(boundary)],
        driver=EasyScipyOptimizeDriver(disp=False),
        ext_vars={'h': 0})
    cost0, _, _ = tf.optimize(state={'h': 0})
    cost10, _, _ = tf.optimize(state={'h': 10})
    npt.assert_almost_equal(cost10, cost0 - 10)


def testCostModelComponentAdditionalOutput():
    def aep_cost(x, y):
        opt_x, opt_y = optimal.T
        return -np.sum((x - opt_x)**2 + (y - opt_y)**2), {'add_out': sum(x)}
    tf = get_tf(AEPCostModelComponent(['x', 'y'], 4, aep_cost, aep_gradients, additional_output=[('add_out', 0)]))
    _, state, _ = tf.optimize()
    npt.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 5)
    npt.assert_equal(sum(state['x']), state['add_out'])



import importlib
import os
import pkgutil
import warnings
import mock
import pytest
import topfarm
import matplotlib.pyplot as plt
import sys


def get_main_modules():
    package = topfarm
    modules = []
    for _, modname, _ in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if 'Colonel' in modname:
                continue
            m = importlib.import_module(modname)

        if 'main' in dir(m):
            modules.append(m)
    return modules


def print_main_modules():
    print("\n".join([m.__name__ for m in get_main_modules()]))


@pytest.mark.parametrize("module", get_main_modules())
def test_main(module):
    # check that all main module examples run without errors
    if os.name == 'posix' and "DISPLAY" not in os.environ:
        pytest.xfail("No display")

    def no_show(*args, **kwargs):
        pass
    plt.show = no_show  # disable plt show that requires the user to close the plot

    def no_print(s, t=None):
        pass

    try:
        with mock.patch.object(module, "__name__", "__main__"):
            with mock.patch.object(module, "print", no_print):
                getattr(module, 'main')()
    except Exception as e:
        raise type(e)(str(e) + ' in %s.main' % module.__name__).with_traceback(sys.exc_info()[2])


if __name__ == '__main__':
    print_main_modules()



import numpy as np
from openmdao.api import ExplicitComponent
from scipy.spatial import ConvexHull
import sys

# ==============================================================================
# This module is deprecated use topfarm.constraint_components.boundary instead
# ==============================================================================


def BoundaryComp(n_wt, xy_boundary, z_boundary=None, xy_boundary_type='convex_hull'):
    if xy_boundary_type == 'polygon':
        return PolygonBoundaryComp(n_wt, xy_boundary, z_boundary)
    else:
        return ConvexBoundaryComp(n_wt, xy_boundary, z_boundary, xy_boundary_type)


class BoundaryBaseComp(ExplicitComponent):
    def __init__(self, n_wt, xy_boundary=None, z_boundary=None, **kwargs):
        sys.stderr.write("%s is deprecated. Use BoundaryConstraint from topfarm.constraint_components.boundary instead\n" % self.__class__.__name__)

        ExplicitComponent.__init__(self, **kwargs)
        self.n_wt = n_wt
        if xy_boundary is None:
            self.xy_boundary = np.zeros((0, 2))
        else:
            self.xy_boundary = np.array(xy_boundary)
        if z_boundary is None:
            z_boundary = []
        if len(z_boundary) > 0:
            z_boundary = np.asarray(z_boundary)
            assert z_boundary.shape[-1] == 2
            if len(z_boundary.shape) == 1:
                z_boundary = np.zeros((self.n_wt, 2)) + [z_boundary]
            assert z_boundary.shape == (self.n_wt, 2)
            assert np.all(z_boundary[:, 0] < z_boundary[:, 1])
        self.z_boundary = z_boundary

#     def setup_as_constraints(self, problem):
#         if len(self.xy_boundary) > 0:
#             problem.model.add_subsystem('xy_bound_comp', self, promotes=['*'])
#             problem.model.add_constraint('boundaryDistances', lower=np.zeros(self.nVertices * self.n_wt))
#         if len(self.z_boundary):
#             problem.model.add_constraint(topfarm.z_key, lower=self.z_boundary[:, 0], upper=self.z_boundary[:, 1])
#
#     def setup_as_penalty(self, problem, penalty=1e10):
#         if len(self.xy_boundary) == 0 and len(self.z_boundary) == 0:
#             return  # no boundary or hub-height constraints
#
#         if len(self.xy_boundary) > 0:
#             subsystem_order = [ss.name for ss in problem.model._static_subsystems_allprocs]
#             problem.model.add_subsystem('xy_bound_comp', self, promotes=['*'])
#             subsystem_order.insert(subsystem_order.index('cost_comp'), 'xy_bound_comp')
#             problem.model.set_order(subsystem_order)
#
#             def xy_boundary_penalty(inputs):
#                 return -np.minimum(inputs['boundaryDistances'], 0).sum()
#         else:
#             def xy_boundary_penalty(inputs):
#                 return 0
#
#         if len(self.z_boundary):
#             def z_boundary_penalty(inputs):
#                 return -np.minimum(inputs[topfarm.z_key] - self.z_boundary[:, 0], 0).sum() + np.maximum(inputs[topfarm.z_key] - self.z_boundary[:, 1], 0).sum()
#         else:
#             def z_boundary_penalty(inputs):
#                 return 0
#
#         self._cost_comp = problem.cost_comp
#         self._org_setup = self._cost_comp.setup
#         self._org_compute = self._cost_comp.compute
#
#         def new_setup():
#             self._org_setup()
#             if len(self.xy_boundary) > 0:
#                 self._cost_comp.add_input('boundaryDistances', val=self.zeros)
#
#         self._cost_comp.setup = new_setup
#
#         def new_compute(inputs, outputs):
#             p = xy_boundary_penalty(inputs) + z_boundary_penalty(inputs)
#             if p == 0:
#                 self._org_compute(inputs, outputs)
#             else:
#                 outputs['cost'] = penalty + p
#         self._cost_comp.compute = new_compute
#         problem._mode = 'rev'


class ConvexBoundaryComp(BoundaryBaseComp):
    def __init__(self, n_wt, xy_boundary=None, z_boundary=None, xy_boundary_type='convex_hull'):
        super().__init__(n_wt, xy_boundary, z_boundary)
        if len(self.xy_boundary):
            self.boundary_type = xy_boundary_type
            self.calculate_boundary_and_normals()
            self.calculate_gradients()
            self.zeros = np.zeros([self.n_wt, self.nVertices])
        else:
            self.zeros = np.zeros([self.n_wt, 0])

    def calculate_boundary_and_normals(self):
        if self.boundary_type == 'convex_hull':
            # find the points that actually comprise a convex hull
            hull = ConvexHull(list(self.xy_boundary))

            # keep only xy_vertices that actually comprise a convex hull and arrange in CCW order
            self.xy_boundary = self.xy_boundary[hull.vertices]
        elif self.boundary_type == 'square':
            min_ = self.xy_boundary.min(0)
            max_ = self.xy_boundary.max(0)
            range_ = (max_ - min_)
            x_c, y_c = min_ + range_ / 2
            r = range_.max() / 2
            self.xy_boundary = np.array([(x_c - r, y_c - r), (x_c + r, y_c - r), (x_c + r, y_c + r), (x_c - r, y_c + r)])
        elif self.boundary_type == 'rectangle':
            min_ = self.xy_boundary.min(0)
            max_ = self.xy_boundary.max(0)
            range_ = (max_ - min_)
            x_c, y_c = min_ + range_ / 2
            r = range_ / 2
            self.xy_boundary = np.array([(x_c - r[0], y_c - r[1]), (x_c + r[0], y_c - r[1]), (x_c + r[0], y_c + r[1]), (x_c - r[0], y_c + r[1])])
        else:
            raise NotImplementedError("Boundary type '%s' is not implemented" % self.boundary_type)

        # get the real number of xy_vertices
        self.nVertices = self.xy_boundary.shape[0]

        # initialize normals array
        unit_normals = np.zeros([self.nVertices, 2])

        # determine if point is inside or outside of each face, and distances from each face
        for j in range(0, self.nVertices):

            # calculate the unit normal vector of the current face (taking points CCW)
            if j < self.nVertices - 1:  # all but the set of point that close the shape
                normal = np.array([self.xy_boundary[j + 1, 1] - self.xy_boundary[j, 1],
                                   -(self.xy_boundary[j + 1, 0] - self.xy_boundary[j, 0])])
                unit_normals[j] = normal / np.linalg.norm(normal)
            else:   # the set of points that close the shape
                normal = np.array([self.xy_boundary[0, 1] - self.xy_boundary[j, 1],
                                   -(self.xy_boundary[0, 0] - self.xy_boundary[j, 0])])
                unit_normals[j] = normal / np.linalg.norm(normal)

        self.unit_normals = unit_normals

    def calculate_gradients(self):
        unit_normals = self.unit_normals

        # initialize array to hold distances from each point to each face
        dfaceDistance_dx = np.zeros([self.n_wt * self.nVertices, self.n_wt])
        dfaceDistance_dy = np.zeros([self.n_wt * self.nVertices, self.n_wt])

        for i in range(0, self.n_wt):
            # determine if point is inside or outside of each face, and distances from each face
            for j in range(0, self.nVertices):

                # define the derivative vectors from the point of interest to the first point of the face
                dpa_dx = np.array([-1.0, 0.0])
                dpa_dy = np.array([0.0, -1.0])

                # find perpendicular distances derivatives from point to current surface (vector projection)
                ddistanceVec_dx = np.vdot(dpa_dx, unit_normals[j]) * unit_normals[j]
                ddistanceVec_dy = np.vdot(dpa_dy, unit_normals[j]) * unit_normals[j]

                # calculate derivatives for the sign of perpendicular distances from point to current face
                dfaceDistance_dx[i * self.nVertices + j, i] = np.vdot(ddistanceVec_dx, unit_normals[j])
                dfaceDistance_dy[i * self.nVertices + j, i] = np.vdot(ddistanceVec_dy, unit_normals[j])

        # return Jacobian dict
        self.dfaceDistance_dx = dfaceDistance_dx
        self.dfaceDistance_dy = dfaceDistance_dy

    def calculate_distance_to_boundary(self, points):
        """
        :param points: points that you want to calculate the distances from to the faces of the convex hull
        :return face_distace: signed perpendicular distances from each point to each face; + is inside
        """

        nPoints = np.array(points).shape[0]
        nVertices = self.xy_boundary.shape[0]
        vertices = self.xy_boundary
        unit_normals = self.unit_normals
        # initialize array to hold distances from each point to each face
        face_distance = np.zeros([nPoints, nVertices])

        # loop through points and find distances to each face
        for i in range(0, nPoints):

            # determine if point is inside or outside of each face, and distances from each face
            for j in range(0, nVertices):

                # define the vector from the point of interest to the first point of the face
                pa = np.array([vertices[j, 0] - points[i, 0], vertices[j, 1] - points[i, 1]])

                # find perpendicular distances from point to current surface (vector projection)
                d_vec = np.vdot(pa, unit_normals[j]) * unit_normals[j]

                # calculate the sign of perpendicular distances from point to current face (+ is inside, - is outside)
                face_distance[i, j] = np.vdot(d_vec, unit_normals[j])
        # print (face_distance)
        return face_distance

#     def setup(self):
#
#         # Explicitly size input arrays
#         self.add_input(topfarm.x_key, np.zeros(self.n_wt), units='m',
#                        desc='x coordinates of turbines in global ref. frame')
#         self.add_input(topfarm.y_key, np.zeros(self.n_wt), units='m',
#                        desc='y coordinates of turbines in global ref. frame')
#
#         # Explicitly size output array
#         # (vector with positive elements if turbines outside of hull)
#         self.add_output('boundaryDistances', self.zeros,
#                         desc="signed perpendicular distances from each turbine to each face CCW; + is inside")
#
#         self.declare_partials('boundaryDistances', [topfarm.x_key, topfarm.y_key])
#         # self.declare_partials('boundaryDistances', ['boundaryVertices', 'boundaryNormals'], method='fd')

    def distances(self, turbineX, turbineY):
        return self.calculate_distance_to_boundary(np.array([turbineX, turbineY]).T)

    def gradients(self, turbineX, turbineY):
        return self.dfaceDistance_dx, self.dfaceDistance_dy

#     def compute(self, inputs, outputs):
#         # calculate distances from each point to each face
#         outputs['boundaryDistances'] = self.distances(**inputs)
#
#     def compute_partials(self, inputs, partials):
#         # return Jacobian dict
#         dx, dy = self.gradients(**inputs)
#
#         partials['boundaryDistances', topfarm.x_key] = dx
#         partials['boundaryDistances', topfarm.y_key] = dy

    def move_inside(self, turbineX, turbineY, turbineZ, pad=1.1):
        x, y, z = [np.asarray(xyz, dtype=np.float) for xyz in [turbineX, turbineY, turbineZ]]
        dist = self.distances(turbineX, turbineY)
        dx, dy = self.gradients(x, y)  # independent of position
        dx = dx[:self.nVertices, 0]
        dy = dy[:self.nVertices, 0]
        for i in np.where(dist.min(1) < 0)[0]:  # loop over turbines that violate edges
            # find smallest movement that where the constraints are satisfied
            d = dist[i]
            v = np.linspace(-np.abs(d.min()), np.abs(d.min()), 100)
            X, Y = np.meshgrid(v, v)
            m = np.ones_like(X)
            for j in range(3):
                m = np.logical_and(m, X * dx[j] + Y * dy[j] >= -dist[i][j])
            index = np.argmin(X[m]**2 + Y[m]**2)
            x[i] += X[m][index]
            y[i] += Y[m][index]
        return x, y, z


class PolygonBoundaryComp(BoundaryBaseComp):
    def __init__(self, n_wt, xy_boundary=None, z_boundary=None, **kwargs):
        BoundaryBaseComp.__init__(self, n_wt, xy_boundary=xy_boundary, z_boundary=z_boundary, **kwargs)

        self.nTurbines = n_wt
        self.zeros = np.zeros(self.nTurbines)
        vertices = self.xy_boundary
        self.nVertices = vertices.shape[0]

        def edges_counter_clockwise(vertices):
            if np.any(vertices[0] != vertices[-1]):
                vertices = np.r_[vertices, vertices[:1]]

            x1, y1 = vertices[:-1].T
            x2, y2 = vertices[1:].T
            double_area = np.sum((x1 - x2) * (y1 + y2))  # 2 x Area (+: counterclockwise
            assert double_area != 0, "Area must be non-zero"
            if double_area < 0:  #
                return edges_counter_clockwise(vertices[::-1])
            else:
                return vertices[:-1], x1, y1, x2, y2

        self.xy_boundary, self.x1, self.y1, self.x2, self.y2 = edges_counter_clockwise(vertices)
        self.min_x, self.min_y = np.min([self.x1, self.x2], 0), np.min([self.y1, self.y2], )
        self.max_x, self.max_y = np.max([self.x1, self.x2], 1), np.max([self.y1, self.y2], 0)
        self.dx = self.x2 - self.x1
        self.dy = self.y2 - self.y1
        self.x2y1 = self.x2 * self.y1
        self.y2x1 = self.y2 * self.x1
        self.length = ((self.y2 - self.y1)**2 + (self.x2 - self.x1)**2)**0.5
        self.edge_unit_vec = (np.array([self.dy, -self.dx]) / self.length)
        v = np.hstack((self.edge_unit_vec, self.edge_unit_vec[:, :1]))
        self.xy2_vec = v[:, :-1] + v[:, 1:]
        self.xy1_vec = np.hstack((self.xy2_vec[:, -1:], self.xy2_vec[:, 1:]))

        self.dEdgeDist_dx = -self.dy / self.length
        self.dEdgeDist_dy = self.dx / self.length
        self._cache_input = None
        self._cache_output = None

#     def setup(self):
#
#         # Explicitly size input arrays
#         self.add_input(topfarm.x_key, np.zeros(self.nTurbines), units='m',
#                        desc='x coordinates of turbines in global ref. frame')
#         self.add_input(topfarm.y_key, np.zeros(self.nTurbines), units='m',
#                        desc='y coordinates of turbines in global ref. frame')
#
#         # Explicitly size output array
#         # (vector with positive elements if turbines outside of hull)
#         self.add_output('boundaryDistances', self.zeros,
#                         desc="signed perpendicular distances from each turbine to each face CCW; + is inside")
#
#         self.declare_partials('boundaryDistances', [topfarm.x_key, topfarm.y_key])
#         # self.declare_partials('boundaryDistances', ['boundaryVertices', 'boundaryNormals'], method='fd')

    def calc_distance_and_gradients(self, x, y):
        """
        distances point(x,y) to edge((x1,y1)->(x2,y2))
        +/-: inside/outside
        case (x,y) closest to edge:
            distances = edge_unit_vec dot (x1-x,y1-y)
            ddist_dx = -(y2-y2)/|edge|
            ddist_dy = (x2-x2)/|edge|
        case (x,y) closest to (x1,y1) (and (x2,y2)):
            sign = sign of distances to nearest edge
            distances = sign * (x1-x^2 + y1-y)^2)^.5
            ddist_dx = sign * 2*x-2*x1 / (2 * distances^.5)
            ddist_dy = sign * 2*y-2*y1 / (2 * distances^.5)
        """
        if np.all(np.array([x, y]) == self._cache_input):
            return self._cache_output

        X, Y = [np.tile(xy, (len(self.x1), 1)).T for xy in [x, y]]  # dim = (ntb, nEdges)
        X1, Y1, X2, Y2, ddist_dX, ddist_dY = [np.tile(xy, (len(x), 1))
                                              for xy in [self.x1, self.y1, self.x2, self.y2, self.dEdgeDist_dx, self.dEdgeDist_dy]]

        # perpendicular distances to edge (dot product)
        d12 = (self.x1 - X) * self.edge_unit_vec[0] + (self.y1 - Y) * self.edge_unit_vec[1]

        # nearest point on edge
        px = X + d12 * self.edge_unit_vec[0]
        py = Y + d12 * self.edge_unit_vec[1]

        # distances to start and end points
        d1 = np.sqrt((self.x1 - X)**2 + (self.y1 - Y)**2)
        d2 = np.sqrt((self.x2 - X)**2 + (self.y2 - Y)**2)

        # use start or end point if nearest point is outside edge
        use_xy1 = (((self.dx != 0) & (px < self.x1) & (self.x1 < self.x2)) |
                   ((self.dx != 0) & (px > self.x1) & (self.x1 > self.x2)) |
                   ((self.dx == 0) & (py < self.y1) & (self.y1 < self.y2)) |
                   ((self.dx == 0) & (py > self.y1) & (self.y1 > self.y2)))
        use_xy2 = (((self.dx != 0) & (px > self.x2) & (self.x2 > self.x1)) |
                   ((self.dx != 0) & (px < self.x2) & (self.x2 < self.x1)) |
                   ((self.dx == 0) & (py > self.y2) & (self.y2 > self.y1)) |
                   ((self.dx == 0) & (py < self.y2) & (self.y2 < self.y1)))

        px[use_xy1] = X1[use_xy1]
        py[use_xy1] = Y1[use_xy1]
        px[use_xy2] = X2[use_xy2]
        py[use_xy2] = Y2[use_xy2]

        distance = d12.copy()
        v = (px[use_xy1] - X[use_xy1]) * self.xy1_vec[0, np.where(use_xy1)[1]] + (py[use_xy1] - Y[use_xy1]) * self.xy1_vec[1, np.where(use_xy1)[1]]
        sign_use_xy1 = np.choose(v >= 0, [-1, 1])
        v = (px[use_xy2] - X[use_xy2]) * self.xy2_vec[0, np.where(use_xy2)[1]] + (py[use_xy2] - Y[use_xy2]) * self.xy2_vec[1, np.where(use_xy2)[1]]
        sign_use_xy2 = np.choose(v >= 0, [-1, 1])

        d12[use_xy2]
        d12[:, 1:][use_xy2[:, :-1]]

        distance[use_xy1] = sign_use_xy1 * d1[use_xy1]
        distance[use_xy2] = sign_use_xy2 * d2[use_xy2]

        length = np.sqrt((X1[use_xy1] - X[use_xy1])**2 + (Y1[use_xy1] - Y[use_xy1])**2)
        ddist_dX[use_xy1] = sign_use_xy1 * (2 * X[use_xy1] - 2 * X1[use_xy1]) / (2 * length)
        ddist_dY[use_xy1] = sign_use_xy1 * (2 * Y[use_xy1] - 2 * Y1[use_xy1]) / (2 * length)

        length = np.sqrt((X2[use_xy2] - X[use_xy2])**2 + (Y2[use_xy2] - Y[use_xy2])**2)
        ddist_dX[use_xy2] = sign_use_xy2 * (2 * X[use_xy2] - 2 * X2[use_xy2]) / (2 * length)
        ddist_dY[use_xy2] = sign_use_xy2 * (2 * Y[use_xy2] - 2 * Y2[use_xy2]) / (2 * length)

        closest_edge_index = np.argmin(np.abs(distance), 1)

        self._cache_input = np.array([x, y])
        self._cache_output = [np.choose(closest_edge_index, v.T) for v in [distance, ddist_dX, ddist_dY]]
        return self._cache_output

    def distances(self, turbineX, turbineY):
        return self.calc_distance_and_gradients(turbineX, turbineY)[0]

    def gradients(self, turbineX, turbineY):
        _, dx, dy = self.calc_distance_and_gradients(turbineX, turbineY)
        return np.diagflat(dx), np.diagflat(dy)

    def move_inside(self, turbineX, turbineY, turbineZ, pad=1.1):
        x, y, z = [np.asarray(xyz, dtype=np.float) for xyz in [turbineX, turbineY, turbineZ]]
        dist = self.distances(turbineX, turbineY)
        dx, dy = map(np.diag, self.gradients(x, y))
        m = dist < 0
        x[m] -= dx[m] * dist[m] * pad
        y[m] -= dy[m] * dist[m] * pad
        return x, y, z



import numpy as np
from topfarm.cost_models.fused_wake_wrappers import FusedWakeGCLWakeModel
from topfarm.cost_models.utils.wind_resource import WindResource
from topfarm.cost_models.cost_model_wrappers import AEPCostModelComponent
from topfarm.tests.test_files import testfilepath
import warnings


class AEPCalculator(object):

    def __init__(self, wind_resource, wake_model, wdir=np.arange(360), wsp=np.arange(3, 25)):
        """
        wind_resource: f(turbine_positions, wdir, wsp) -> WD[nWT,nWdir,nWsp], WS[nWT,nWdir,nWsp], TI[nWT,nWdir,nWsp), Weight[nWdir,nWsp]
        wake_model: f(turbine_positions, WD[nWT,nWdir,nWsp], WS[nWT,nWdir,nWsp], TI[nWT,nWdir,nWsp) -> power[nWdir,nWsp] (W)
        """
        self.wind_resource = wind_resource
        self.wake_model = wake_model
        self.wdir = wdir
        self.wsp = wsp

    def __call__(self, x, y, **_):
        turbine_positions = np.array([x, y]).T
        no_wake_WD, no_wake_WS, no_wake_TI, weight = self.wind_resource(turbine_positions, self.wdir, self.wsp)
        power_GW = self.wake_model(turbine_positions, no_wake_WD, no_wake_WS, no_wake_TI) / 1e9
        return np.sum(power_GW * weight) * 24 * 365

    def get_TopFarm_cost_component(self):
        n_wt = self.wake_model.windFarm.nWT

        def aep_func(x, y, **kwargs):
            return AEPCalculator(self.wind_resource, self.wake_model)(x, y).sum()

        return AEPCostModelComponent(
            input_keys=['x', 'y'],
            n_wt=n_wt,
            cost_function=aep_func,
            output_unit='GWh')


def main():
    if __name__ == '__main__':
        f = [0.035972, 0.039487, 0.051674, 0.070002, 0.083645, 0.064348,
             0.086432, 0.117705, 0.151576, 0.147379, 0.10012, 0.05166]
        A = [9.176929, 9.782334, 9.531809, 9.909545, 10.04269, 9.593921,
             9.584007, 10.51499, 11.39895, 11.68746, 11.63732, 10.08803]
        k = [2.392578, 2.447266, 2.412109, 2.591797, 2.755859, 2.595703,
             2.583984, 2.548828, 2.470703, 2.607422, 2.626953, 2.326172]
        wr = WindResource(np.array(f) / 100, A, k, ti=np.zeros_like(f) + .1)
        wf_3tb = testfilepath + "wind_farms/3tb.yml"
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')  # temporarily disable fusedwake warnings
            wm = FusedWakeGCLWakeModel(wf_3tb)
            aep_calc = AEPCalculator(wr, wm)
            print(aep_calc(wm.windFarm.pos[:, 0], wm.windFarm.pos[:, 1]))


main()



import unittest

import numpy as np
from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
from topfarm import TopFarm

from topfarm.plotting import NoPlot
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.constraint_components.boundary import XYBoundaryConstraint,\
    PolygonBoundaryComp
from topfarm._topfarm import TopFarmProblem


def get_tf(initial, optimal, boundary, plot_comp=NoPlot()):
    initial, optimal = map(np.array, [initial, optimal])
    return TopFarmProblem(
        {'x': initial[:, 0], 'y': initial[:, 1]},
        DummyCost(optimal),
        constraints=[XYBoundaryConstraint(boundary, 'polygon')],
        driver=EasyScipyOptimizeDriver(tol=1e-8, disp=False),
        plot_comp=plot_comp)


def testPolygon():
    boundary = [(0, 0), (1, 1), (2, 0), (2, 2), (0, 2)]
    b = PolygonBoundaryComp(0, boundary)
    np.testing.assert_array_equal(b.xy_boundary[:, :2], [[0, 0],
                                                         [1, 1],
                                                         [2, 0],
                                                         [2, 2],
                                                         [0, 2],
                                                         [0, 0]])


def testPolygonConcave():
    optimal = [(1.5, 1.3), (4, 1)]
    boundary = [(0, 0), (5, 0), (5, 2), (3, 2), (3, 1), (2, 1), (2, 2), (0, 2), (0, 0)]
    plot_comp = NoPlot()  # DummyCostPlotComp(optimal)
    initial = [(-0, .1), (4, 1.5)][::-1]
    tf = get_tf(initial, optimal, boundary, plot_comp)
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal, 4)
    plot_comp.show()


def testPolygonTwoRegionsStartInWrong():
    optimal = [(1, 1), (4, 1)]
    boundary = [(0, 0), (5, 0), (5, 2), (3, 2), (3, 0), (2, 0), (2, 2), (0, 2), (0, 0)]
    plot_comp = NoPlot()  # DummyCostPlotComp(optimal, delay=.1)
    initial = [(3.5, 1.5), (0.5, 1.5)]
    tf = get_tf(initial, optimal, boundary, plot_comp)
    tf.optimize()
    plot_comp.show()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal, 4)



from topfarm.cost_models.dummy import DummyCost
from topfarm.plotting import NoPlot
from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
from topfarm.tests.test_files import xy3tb
from topfarm.constraint_components.spacing import SpacingConstraint, SpacingComp
from topfarm.tests import npt
from openmdao.utils.assert_utils import assert_check_partials
from openmdao.api import Problem, IndepVarComp
import numpy as np


def test_spacing():
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(2)])
    tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tf.plot_comp.show()
    tol = 1e-4
    assert sum((tb_pos[2] - tb_pos[0])**2) > 2**2 - tol  # check min spacing


def test_spacing_as_penalty():
    driver = SimpleGADriver()
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(2)],
                      driver=driver)

    # check normal result if spacing constraint is satisfied
    assert tf.evaluate()[0] == 45
    # check penalized result if spacing constraint is not satisfied
    assert tf.evaluate({'x': [3, 7, 4.], 'y': [-3., -7., -3.], 'z': [0., 0., 0.]})[0] == 1e10 + 3


def test_satisfy():
    sc = SpacingComp(n_wt=3, min_spacing=2)
    state = sc.satisfy(dict(zip('xy', xy3tb.desired.T)))
    x, y = state['x'], state['y']
    npt.assert_array_less(y, x)


def test_satisfy2():
    n_wt = 5
    sc = SpacingComp(n_wt=n_wt, min_spacing=2)
    theta = np.linspace(0, 2 * np.pi, n_wt, endpoint=False)
    x0, y0 = np.cos(theta), np.sin(theta)

    state = sc.satisfy({'x': x0, 'y': y0})
    x1, y1 = state['x'], state['y']
    if 0:
        import matplotlib.pyplot as plt
        colors = ['b', 'r', 'm', 'c', 'g', 'y', 'orange', 'indigo', 'grey']
        for i, (x0_, y0_, x1_, y1_) in enumerate(zip(x0, y0, x1, y1)):
            c = colors[i]
            plt.plot([x0_], [y0_], '>', color=c)
            plt.plot([x0_, x1_], [y0_, y1_], '-', color=c, label=i)
            plt.plot([x1_], [y1_], '.', color=c)
        plt.axis('equal')
        plt.legend()
        plt.show()

    dist = np.sqrt(sc._compute(x1, y1))
    npt.assert_array_less(2, dist)


def test_partials():
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(2)])
    # if complex numbers work: uncomment tf.setup below and
    # change method='cs' and step=1e-40 in check_partials
    # tf.setup(force_alloc_complex=True)

    # run to get rid of zeros initializaiton, otherwise not accurate partials
    tf.run_model()
    check = tf.check_partials(compact_print=True,
                              includes='spacing*',
                              method='fd',
                              step=1e-6,
                              form='central')
    atol = 1.e-6
    rtol = 1.e-6
    try:
        assert_check_partials(check, atol, rtol)
    except ValueError as err:
        print(str(err))
        raise


def test_partials_many_turbines():
    n_wt = 10
    theta = np.linspace(0, 2 * np.pi, n_wt, endpoint=False)
    sc = SpacingComp(n_wt=n_wt, min_spacing=2, const_id="")
    tf = Problem()
    ivc = IndepVarComp()
    ivc.add_output('x', val=np.cos(theta))
    ivc.add_output('y', val=np.sin(theta))
    tf.model.add_subsystem('ivc', ivc, promotes=['*'])
    tf.model.add_subsystem('sc', sc, promotes=['x', 'y', 'wtSeparationSquared'])
    tf.setup()
    tf.run_model()

    check = tf.check_partials(compact_print=True,
                              includes='sc*',
                              method='fd',
                              step=1e-6,
                              form='central')

    fil = {'sc': {key: val for key, val in check['sc'].items()
                  if 'penalty' not in key[0]}}

    atol = 1.e-6
    rtol = 1.e-6
    try:
        assert_check_partials(fil, atol, rtol)
    except ValueError as err:
        print(str(err))
        raise


if __name__ == '__main__':
    test_satisfy2()



'''
Created on 17. jul. 2018

@author: mmpe
'''
from topfarm.utils import smart_start
import numpy as np
from topfarm.tests import npt
from topfarm import TopFarmProblem

from py_wake.examples.data import hornsrev1
from py_wake.wake_models.noj import NOJ
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.constraint_components.boundary import XYBoundaryConstraint,\
    CircleBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.cost_models.py_wake_wrapper import PyWakeAEP
from py_wake.examples.data.iea37._iea37 import IEA37Site
from scipy.interpolate.interpolate import RegularGridInterpolator
from py_wake.site._site import UniformSite


def tests_smart_start():
    xs_ref = [1.6, 14.1, 1.6, 7.9, 14.1, 7.9, 19.9, 19.9, 7.8,
              5.8, 14.2, 5.8, 1.5, 16.2, 16.2, 1.6, 3.7, 14.1, 7.9, 3.7]
    ys_ref = [1.6, 1.6, 7.9, 1.6, 7.9, 7.9, 1.6, 7.9, 5.8, 7.8, 5.8, 1.5, 5.8, 7.8, 1.5, 3.7, 1.6, 3.7, 3.7, 7.9]

    x = np.arange(0, 20, 0.1)
    y = np.arange(0, 10, 0.1)
    YY, XX = np.meshgrid(y, x)
    val = np.sin(XX) + np.sin(YY)
    N_WT = 20
    min_space = 2.1
    np.random.seed(0)
    xs, ys = smart_start(XX, YY, val, N_WT, min_space)
    npt.assert_array_almost_equal([xs, ys], [xs_ref, ys_ref])

    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val, 100)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        print(np.round(xs, 1).tolist())
        print(np.round(ys, 1).tolist())

        plt.axis('equal')
        plt.show()


def tests_smart_start_no_feasible():
    x = np.arange(0, 20, 0.1)
    y = np.arange(0, 10, 0.1)
    YY, XX = np.meshgrid(y, x)
    val = np.sin(XX) + np.sin(YY)
    N_WT = 20
    min_space = 5.1

    xs, ys = smart_start(XX, YY, val, N_WT, min_space)

    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val, 100)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        plt.axis('equal')
        plt.show()
        print(xs)
    assert np.isnan(xs).sum() == 12


def test_smart_start_aep_map():
    site = IEA37Site(16)
    n_wt = 4
    np.random.seed(1)
    x, y = site.initial_position[:n_wt].T
    wd_lst = np.arange(0, 360, 45)
    ws_lst = [10]
    turbines = hornsrev1.HornsrevV80()
    site = UniformSite([1], .75)
    site.default_ws = ws_lst
    site.default_wd = wd_lst

    aep = PyWakeAEP(wake_model=NOJ(site, turbines))
    aep_1wt = aep.calculate_AEP([0], [0]).sum()

    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep.get_TopFarm_cost_component(n_wt),
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingConstraint(160), CircleBoundaryConstraint((0, 0), 500)]
    )
    x = np.arange(-500, 500, 10)
    y = np.arange(-500, 500, 10)
    XX, YY = np.meshgrid(x, y)

    tf.smart_start(XX, YY, aep.get_aep4smart_start(wd=wd_lst, ws=ws_lst), radius=40)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        for i, _ in enumerate(wt_x, 1):
            print(aep.calculate_AEP(wt_x[:i], wt_y[:i]).sum((1, 2)))
        X_j, Y_j, aep_map = aep.aep_map(x, y, 0, wt_x, wt_y, ws=ws_lst, wd=wd_lst)
        print(tf.evaluate())
        import matplotlib.pyplot as plt
        c = plt.contourf(X_j, Y_j, aep_map, 100)
        plt.colorbar(c)
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_almost_equal(aep_1wt * n_wt, tf['AEP'], 5)


def test_smart_start_aep_map_large_radius():
    site = IEA37Site(16)
    n_wt = 4
    np.random.seed(0)
    x, y = site.initial_position[:n_wt].T
    wd_lst = np.arange(0, 360, 45)
    ws_lst = [10]
    turbines = hornsrev1.HornsrevV80()
    site = UniformSite([1], .75)
    site.default_ws = ws_lst
    site.default_wd = wd_lst

    aep = PyWakeAEP(wake_model=NOJ(site, turbines))
    aep_1wt = aep.calculate_AEP([0], [0]).sum()
    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep.get_TopFarm_cost_component(n_wt),
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingConstraint(160), CircleBoundaryConstraint((0, 0), 2000)]
    )
    x = np.arange(-2000, 2000, 100)
    y = np.arange(-2000, 2000, 100)
    XX, YY = np.meshgrid(x, y)

    tf.smart_start(XX, YY, aep.get_aep4smart_start(wd=wd_lst, ws=ws_lst), radius=40)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        for i, _ in enumerate(wt_x, 1):
            print(aep.calculate_AEP(wt_x[:i], wt_y[:i]).sum((1, 2)))
        X_j, Y_j, aep_map = aep.aep_map(x, y, 0, wt_x, wt_y, ws=ws_lst, wd=wd_lst)
        print(tf.evaluate())
        import matplotlib.pyplot as plt
        c = plt.contourf(X_j, Y_j, aep_map, 100)
        plt.colorbar(c)
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_almost_equal(aep_1wt * n_wt, tf['AEP'], 3)



from openmdao.drivers.doe_generators import DOEGenerator, ListGenerator
import numpy as np
import copy
import sys


class ConstrainedDiscardGenerator(DOEGenerator):
    def __init__(self, generator):
        DOEGenerator.__init__(self)
        if isinstance(generator, list):
            generator = ListGenerator(generator)
        self.generator = generator

    def __call__(self, design_vars, model=None):
        constr = model._get_subsystem('pre_constraints')
        xy_boundary_comp = constr._get_subsystem('xy_bound_comp')
        spacing_comp = constr._get_subsystem('spacing_comp')
        for xyz_tuple_lst in self.generator(design_vars, model=model):
            x, y = ([np.array(t[1]) for t in xyz_tuple_lst] + [0])[:2]
            if spacing_comp:
                dist = spacing_comp._compute(x, y)
                if np.any(dist < spacing_comp.min_spacing**2):
                    continue

            if xy_boundary_comp:
                dist_to_boundary = xy_boundary_comp.distances(x, y)
                if np.any(dist_to_boundary < 0):
                    continue
            yield xyz_tuple_lst


class ConstrainedGenerator(DOEGenerator):
    def __init__(self, generator, n_iter=1000, step_size=0.1, offset=0.5, verbose=False):
        DOEGenerator.__init__(self)
        self.n_iter = n_iter
        self.step_size = step_size
        self.offset = offset
        if isinstance(generator, list):
            generator = ListGenerator(generator)
        self.generator = generator
        self.verbose = verbose

    def __call__(self, design_vars, model):
        for design_var_tuple_lst in self.generator(design_vars, model=model):
            state = {k.replace("indeps.", ''): np.array(v) for k, v in design_var_tuple_lst}
            for _ in range(100):
                state_before = copy.deepcopy(state)
                for constr_comp in model.constraint_components:
                    constr_comp.satisfy(state)
                if all([np.all(state[k] == state_before[k]) for k in state_before.keys()]):
                    break
            xyz_tuple_lst = [("indeps." + k, v) for k, v in state.items()]
            yield xyz_tuple_lst


class ConstrainedXYZGenerator(ConstrainedGenerator):
    def __init__(self, generator, n_iter=1000, step_size=0.1, offset=0.5, verbose=False):
        super().__init__(generator, n_iter=n_iter, step_size=step_size, offset=offset, verbose=verbose)
        sys.stderr.write("%s is deprecated. Use ConstrainedGenerator instead\n" % self.__class__.__name__)


class ConstrainedDiscardXYZGenerator(ConstrainedDiscardGenerator):
    def __init__(self, generator):
        ConstrainedDiscardGenerator.__init__(self, generator)
        sys.stderr.write("%s is deprecated. Use ConstrainedDiscardGenerator instead\n" % self.__class__.__name__)




import numpy as np
from topfarm.cost_models.utils.wind_resource import WindResource
from topfarm.tests.test_files import testfilepath
from topfarm.cost_models.fused_wake_wrappers import FusedWakeGCLWakeModel
from topfarm.cost_models.utils.aep_calculator import AEPCalculator
import warnings
from topfarm.tests import npt


def test_AEPCalculator():
    f = [1, 0, 0, 0]
    A = [9.176929, 10, 10, 10]
    k = [2.392578, 2, 2, 2]
    wr = WindResource(np.array(f), A, k, ti=np.zeros_like(f) + .1)
    wf_3tb = testfilepath + "wind_farms/3tb.yml"
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        wm = FusedWakeGCLWakeModel(wf_3tb)
        aep_calc = AEPCalculator(wr, wm)
        npt.assert_almost_equal(aep_calc([-1600, 0, 1600], [0, 0, 0]), 22.3178800761)
